// MRCData.cc - Copyright 1997  Steven Ludtke
// This is a generic EM data object extracted from 'emtool'
// It is not well documented at this point

#include  "MRCData.h"
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "List.h"
#include <sys/types.h>
#include <time.h>
#include "cwp.h"
#ifdef SGI
#include "bstring.h"
#endif
#include <malloc.h>

int gpf(int val);
float getenvfd(char *name,float def);
float frand(float lo,float hi);
float grand(float mean, float sigma);
void save_data(float *x, float *y, int n, char *fsp);

//#define EECH	1	// debug flag

#define uchar unsigned char
#define ushort unsigned short
#define ulong unsigned long

inline float asub(float a1,float a2)
{
float r;

r=fabs(a1-a2);
if (r>M_PI) r=fabs(r-M_PI*2.0);
return r;
}

void do_progress(int i,int ny,char *label);

void brev(unsigned char *data,int count) 
{
int i;
unsigned char t;

for (i=count*4-4; i>=0; i-=4) {
	t=data[i];
	data[i]=data[i+3];
	data[i+3]=t;

	t=data[i+1];
	data[i+1]=data[i+2];
	data[i+2]=t;
}
return;
}


// this returns a List of MRCData objects with the individual images
// no byte order swapping yet
List *readIMAGIC(char *fsp3)
{
List *list;
MRCData *tmp;
int ret;
IMAGICH hed,*hed2;
FILE *in,*inh;
char *fsp,*fsp2,*t;
char *lab[] = { "","Byte reversed ","" };
int i,count,type,ord=0;

fsp=(char *)malloc(strlen(fsp3)+5);
fsp2=(char *)malloc(strlen(fsp3)+5);
strcpy(fsp,fsp3);
strcpy(fsp2,fsp3);
if (strrchr(fsp3,'.')!=NULL) { *(strrchr(fsp,'.'))=0; *(strrchr(fsp2,'.'))=0; }

strcat(fsp2,".hed");
strcat(fsp,".img");

inh=fopen(fsp2,"r");
if (inh==NULL) { printf("ERROR: cannot open %s\n",fsp2); return NULL; }
in=fopen(fsp,"r");
if (in==NULL) { printf("ERROR: cannot open %s\n",fsp); return 0; }

list=new List;
fread(&hed,4,256,inh);
rewind(inh);

// check byte order of image and machine
t=(char *)&hed.headrec;
if (t[0]==0&&t[1]==0) ord=1;
i=127;
t=(char *)&i;
if (t[0]!=127) ord++;

if (strncmp(hed.type,"REAL",4)==0) type=0;
else if (strncmp(hed.type,"INTG",4)==0) type=1;
else {
	printf("Unknown type %4s.\n",hed.type);
	fclose(in);
	fclose(inh);
	return 0;
}

	if (ord==1) { 
		brev((uchar *)&hed,14);
		brev((uchar *)&hed.ixold,14);
		brev((uchar *)hed.SPACE,207);
	}
//printf("imagic file with %d %dx%d %4s images (cell=%f)\n",
//	hed.count+1,hed.nx,hed.ny,hed.type,hed.cellx);
count=hed.count+1;

for (i=0; i<count; i++) {
	if (count>1) do_progress(i*1000000,count*1000000,"Reading multiple images");
	tmp=new MRCData;
	hed2=(IMAGICH *)calloc(sizeof(IMAGICH),1);
	fread(hed2,4,256,inh);
	if (ord==1) {
		brev((uchar *)hed2,14);
		brev((uchar *)&hed2->ixold,14);
		brev((uchar *)hed2->SPACE,207);
	}
//	printf("Image: %dx%d %s\n",hed.nx,hed.ny,hed.type);
	if (type==0) ret=tmp->setFloatRealFromStream(in ,hed2->nx ,hed2->ny ,1 ,ord);
	else if (type==1) ret=tmp->setShortRealFromStream(in ,hed2->nx ,hed2->ny ,1 ,ord);
	else ret=0; // for safety
	if (strchr(fsp,'/')) tmp->setName(strrchr(fsp,'/')+1);
	tmp->setRenMinMax(tmp->Min() ,tmp->Max());
	tmp->setIMHed(hed2);
	if (ret) list->addObject(tmp);
	else delete tmp;
}
fclose(in);
fclose(inh);

if (count>1) do_progress(-1,0,NULL);
free(fsp);
free(fsp2);
return list;
}

MRCData::MRCData()
{
mrc_header=NULL;
im_header=NULL;
flags=0;
rdata=(float *)malloc(4096);
nx=ny=nz=0;
min=histmin=max=histmax=0;
mean=sigma=edgemean=edgesig=0.0;
mag=getenvfd("mag",30.0);
pixres=getenvfd("scanres",25.0)/mag*10.0;
bzero((char *)hist,257*sizeof(float));
renMin=renMax=-999.0;
name=NULL;
filespec=NULL;
boxes=(sRect *)malloc(sizeof(sRect)*4);
boxesa=4;
nboxes=0;
seqno=0;
distdat=NULL;
lastdist=-1;
pstate=NULL;
}

// this version initializes headers from another data set
MRCData::MRCData(MRCData *src)
{
if (src->getHeader()) {
	mrc_header=(MRC_header *)malloc(sizeof(MRC_header));
	bcopy((char *)src->getHeader(),(char *)mrc_header,sizeof(MRC_header));
}
else mrc_header=NULL;
im_header=NULL;
flags=0;
rdata=(float *)malloc(4096);
nx=ny=nz=0;
min=histmin=max=histmax=0;
mean=sigma=edgemean=edgesig=0.0;
mag=src->Mag();
pixres=src->Pixres();
bzero((char *)hist,257*sizeof(float));
renMin=renMax=-999.0;
name=NULL;
filespec=NULL;
setName(src->Name());
setFile(src->File());
boxes=(sRect *)malloc(sizeof(sRect)*4);
boxesa=4;
nboxes=0;
seqno=0;
pstate=NULL;
distdat=NULL;
lastdist=-1;
}

MRCData::~MRCData()
{
if (mrc_header) free(mrc_header);
if (im_header) free(im_header);
if (rdata) free(rdata);
if (name) free(name);
if (filespec) free(filespec);
free(boxes);
if (distdat) free(distdat);
if (pstate) free(pstate);
}

char * MRCData::File() 
{
if (filespec) return filespec; 
return "unknown";
}

int MRCData::setFile( char *newname)
{
if (filespec) free(filespec);
if (!newname || strcmp(newname,"unknown")==0) { filespec=NULL; return 0; }
filespec=(char *)malloc(strlen(newname)+1);
strcpy(filespec,newname);
return 1;
}

void MRCData::setIMHed(IMAGICH *hed)
{
im_header=hed;
return;
}

void MRCData::setMRCHed(MRC_header *hed)
{
mrc_header=hed;
return;
}

char *MRCData::Name() { if (name) return name; return "na"; }

int MRCData::setName( char * newname)
{
if (name) free(name);
if (!newname || strcmp(newname,"na")==0) { name=NULL; return 1; }
name=(char *)malloc(strlen(newname)+1);
strcpy(name,newname);
return 1;
}

// reads MRC data from another data set
int MRCData::dataFromData(MRCData *src , int x0 , int y0 , int size)
{
if (!mrc_header) mrc_header=(MRC_header *)malloc(sizeof(MRC_header));
bcopy((char *)src->getHeader(),(char *)mrc_header,sizeof(MRC_header));
setFile(src->File());
nx=ny=size;
rdata=(float *)realloc(rdata,sizeof(float)*nx*ny);
src->extractReal(rdata ,x0 ,y0 ,size ,size);

return 1;
}

int MRCData:: appendImagic( char * file)
{
FILE *out;
char *s;
int i;
time_t itime;
struct tm *tm;
IMAGICH tmp;

if (!im_header) im_header=(IMAGICH *)calloc(sizeof(IMAGICH),1);

itime=time(0);
tm=localtime(&itime);

// header filespec
s=(char *)malloc(strlen(file)+4);
strcpy(s,file);
if (strlen(s)>4 && s[strlen(s)-4]=='.') s[strlen(s)-4]=0;
strcat(s,".hed");

im_header->count=0;
im_header->imnum=1;
// check byte order if file exists
out=fopen(s,"r+");
if (out!=NULL) {
	fread(&tmp,4,256,out);
	    if ((tmp.imnum&0xfff)==0) {
		printf("File exists and has wrong byte order!\n");
		return 0;
	}
	tmp.count++;
	rewind(out);
	fwrite(&tmp,4,256,out);
	im_header->imnum=tmp.count+1;
}
fclose (out);

out=fopen(s,"a+");
if (out==NULL) { printf("Cannot write to %s!\n",s); return 0; }
if (name) strcpy(im_header->label,name);
im_header->error=0;
im_header->headrec=1;
im_header->mday=tm->tm_mday;
im_header->month=tm->tm_mon;
im_header->year=tm->tm_year+1900;
im_header->hour=tm->tm_hour;
im_header->minute=tm->tm_min;
im_header->sec=tm->tm_sec;
im_header->reals=nx*ny;
im_header->pixels=nx*ny;
im_header->ny=ny;
im_header->nx=nx;
im_header->type[0]='R';
im_header->type[1]='E';
im_header->type[2]='A';
im_header->type[3]='L';
im_header->ixold=im_header->iyold=0;
im_header->avdens=mean;
im_header->sigma=0;
im_header->oldav=0;
im_header->max=max;
im_header->min=min;

// buffering parameters IMAGIC seems to need
im_header->lbuf=nx;
im_header->inn=1;
im_header->iblp=ny;
im_header->ifb=0;
im_header->lbr= -1;
im_header->lastlr= -1;
im_header->lastlw=1;
im_header->num=8; // ???
im_header->nhalf=nx/2;
im_header->ibsd=nx*2;
im_header->ihfl=7;
im_header->lcbr= -1;
im_header->lcbw=1;
im_header->imstr= -1;
im_header->imstw= -1;
im_header->istart=1;
im_header->iend=nx;
im_header->leff=nx;
im_header->linbuf=nx*2;
im_header->ntotbuf= -1;
im_header->icstart=1;
im_header->icend=nx/2;

fwrite(im_header,4,256,out);
fclose(out);

s[strlen(s)-4]=0;
strcat(s,".img");
out=fopen(s,"a+");
if (out==NULL) { printf("Cannot write to %s!\n",s); return 0; }
for (i=ny*nz-1; i>=0; i--) fwrite(&rdata[i*nx],4,nx,out);
fclose(out);

return 1;
}

// writes 2d MRC files (if sh=1, write short instead of float)
int MRCData::writeMRCData( char * fsp , int sh)
{
FILE *out;
char s[80],*t;
struct tm *tm;
time_t tim=time(0);
int i;
float a, p;
unsigned short *sdata;
float sscale,sorig;

out=fopen(fsp,"w");
if (out==NULL) { printf("Cannot open file for writing\n"); return 0; }

if (!mrc_header) mrc_header=(MRC_header *)calloc(sizeof(MRC_header),1);

mrc_header->xlen=mrc_header->nx=nx;
mrc_header->ylen=mrc_header->ny=ny;
mrc_header->zlen=mrc_header->nz=nz;
sdata=(unsigned short *)rdata;

if (mrc_header->alpha==0) mrc_header->alpha=mrc_header->beta=mrc_header->gamma=90.0;
if (flags&MRC_COMPLEX) {
	mrc_header->nx/=2;
	if (sh) mrc_header->mode=MODE_short_COMPLEX;
	else mrc_header->mode=MODE_float_COMPLEX;

// convert from amp,phase to r,i
	for (i=0; i<nx*ny*nz; i+=2) {
		a=rdata[i]*cos(rdata[i+1]);
		p=rdata[i]*sin(rdata[i+1]);
		rdata[i]=a;
		rdata[i+1]=p;
	}
}
else {
	if (sh) mrc_header->mode=MODE_short;
	else mrc_header->mode=MODE_float;
}

// if this is a short save, we compact the data in place
if (sh) {
//	if (max<3.5&&min>=0) { sscale=12800.0; sorig=0; }
	sscale=4095.0/(max-min);
	sorig=min;
	for (i=0; i<nx*ny*nz; i++) sdata[i]=(ushort)((rdata[i]-sorig)*sscale);
}

mrc_header->nxstart=mrc_header->nystart=mrc_header->nzstart=0;
mrc_header->mapc=1;
mrc_header->mapr=2;
mrc_header->maps=3;


if (sh) {
	mrc_header->amax=(max-sorig)*sscale;
	mrc_header->amin=(min-sorig)*sscale;
	mrc_header->amean=(mean-sorig)*sscale;
}
else {
	mrc_header->amax=max;
	mrc_header->amin=min;
	mrc_header->amean=mean;
}

if (mrc_header->nlabl<MRC_NUM_LABELS) {
	if ((t=name)==NULL) t="no name";
	tm=localtime(&tim);
	sprintf(mrc_header->labels[mrc_header->nlabl],"emtool: %d/%d/%02d %d:%02d %s",tm->tm_mon+1,tm->tm_mday,tm->tm_year%100,tm->tm_hour,tm->tm_min,t);
	mrc_header->nlabl++;
}

fwrite(mrc_header,sizeof(MRC_header),1,out);
if (sh) fwrite(rdata,nx*sizeof(short),ny*nz,out);
else fwrite(rdata,nx*sizeof(float),ny*nz,out);
fclose(out);

// if this is a short save, we expand the data in place
if (sh) {
	for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=sdata[i]/sscale+sorig;
}

if (flags&MRC_COMPLEX) {
    for (i=0; i<nx*ny*nz; i+=2) {
	a=hypot(rdata[i],rdata[i+1])/nx*ny*nz;
	p=atan2(rdata[i+1],rdata[i]);
	rdata[i]=a;
	rdata[i+1]=p;
    }
}

return 1;
}

int   MRCData::writePGM( char * fsp)
{
FILE *out;
unsigned char *dat;

out=fopen(fsp,"w");
if (out==NULL) { printf("Cannot open file for writing\n"); return 0; }

dat=(unsigned char *)malloc(nx*ny);
renderAmp8(dat,0,0,nx,ny,nx,1.0,0,255);

fprintf(out,"P5\n%d %d\n255\n",nx,ny);
fwrite(dat,nx,ny,out);

fclose(out);
free(dat);

return 1;
}

// reads 2d (&3d?) MRC files
int MRCData::readMRCData( char * fsp)
{
FILE *in;
int i,j,l,pipe=0;
int ord=0;	// 1 if the file order is MSB first
int mord=0; // 1 if the machine order is MSB first
float f,a,p;
char s[800],*m2,w;
unsigned char *cdata,*cdata2,c;
static char *str[] = { "char", "short", "float", "short complex", "float complex" };
static char *str2[] = { "" , "byte reversed" , "" };
static char *str3[] = { "" , "compressed" };

#ifdef EECH
printf("MRC open '%s'\n",fsp);
#endif

if (!mrc_header) mrc_header=(MRC_header *)calloc(sizeof(MRC_header),1);

if (strcmp(&fsp[strlen(fsp)-3],".gz")==0 || strcmp(&fsp[strlen(fsp)-2],".Z")==0) {
	sprintf(s,"zcat %s",fsp);
	in=popen(s,"r");
	pipe=1;
}
else in=fopen(fsp,"r");
if (in==NULL) {
	printf("Cannot open file\n");
	return 0;
}
if (strrchr(fsp,'/')==NULL) {
	name=(char *)malloc(strlen(fsp)+1);
	strcpy(name,fsp);
}
else {
	name=(char *)malloc(strlen(strrchr(fsp,'/'))+1);
	strcpy(name,strrchr(fsp,'/')+1);
}
// setFile(name);
setFile(fsp);

i=1;
if (*((char *)&i)!=1) mord=1;

if (!fread(mrc_header,sizeof(MRC_header),1,in)) return 0;

// This swaps the byte order in the header on appropriate machines
m2=(char *)mrc_header;
if (m2[0]==0 && m2[1]==0) ord=1;
if (mord ^ ord) {
	for (i=0; i<(27+MRC_USER)*4; i+=4) {
		w=m2[i]; m2[i]=m2[i+3]; m2[i+3]=w;
		w=m2[i+1]; m2[i+1]=m2[i+2]; m2[i+2]=w;
	}
}

//printf("ord=%d mord=%d\n",ord,mord);

if (mrc_header->ny==1) { printf("This appears to be a 1d file. I only read 2d files.\n"); return 0; }
if (mrc_header->nz!=1) { printf("This appears to be a 3d file. I only read 2d files.\n"); return 0; }

if (mrc_header->mode==MODE_float_COMPLEX || mrc_header->mode==MODE_short_COMPLEX) 
	mrc_header->nx*=2;
nx=mrc_header->nx;
ny=mrc_header->ny;
nz=mrc_header->nz;

//if (mrc_header->nz==1) printf("%s is a %s %s %dx%d 2D %s MRC File\n",strrchr(fsp,'/')+1,
//str2[ord+mord],str3[pipe],nx,ny,str[mrc_header->mode]);
//else printf("%s id a %dx%dx%d 3D MRC File\n",strrchr(fsp,'/'),nx,ny,nz);

if (mrc_header->nxstart!=0||mrc_header->nystart!=0||mrc_header->nzstart!=0) printf("WARNING: nx/y/z start not zero\n");

//if (header.mapc!=1 || header.mapr!=2 || header.maps!=3) printf("WARNING: map not 1,2,3\n");

if (mrc_header->xlen==0) mrc_header->xlen=1.0;
if (mrc_header->ylen==0) mrc_header->ylen=1.0;
if (mrc_header->zlen==0) mrc_header->zlen=1.0;

// This reads 2d&3d arrays
//if (header.nz==1) {
	rdata=(float *)realloc(rdata,sizeof(float)*nx*ny*nz);
	cdata=(unsigned char *)rdata;

	switch(mrc_header->mode) {
	case MODE_char:
		flags&=~MRC_COMPLEX;
		for (j=0; j<ny; j++) {
			if ((i=fread(&cdata[j*nx*nz],nx,nz,in))!=nz) {
				printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
//				return 0;
			}
			do_progress(j*nx,nx*ny,"Reading 8 bit data...");
		}
		do_progress(nx*ny,nx*ny,"Converting to float");
		for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=(float)cdata[i]/100.0;
		break;
	case MODE_short:
		flags&=~MRC_COMPLEX;
#ifdef EECH
printf("Short file read\n");
#endif	
		for (j=0; j<ny; j++) {
			if ((i=fread(&cdata[j*nx*nz*sizeof(short)],nx*sizeof(short),nz,in))!=nz) {
				printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
				return 0;
			}
			do_progress(j*nx*2,ny*nx*2,"Reading 16 bit data...");
		}
		do_progress(ny*nx*2,ny*nx*2,"Converting to float");
		if (ord) {
			for (i=nx*ny*nz-1; i>=0; i--) 
				rdata[i]=(float)((cdata[i*2]<<8)+cdata[i*2+1]); 
		}
		else {
			for (i=nx*ny*nz-1; i>=0; i--) 
				rdata[i]=(float)((cdata[i*2+1]<<8)+cdata[i*2]); 
		}

#ifdef EECH
printf("Short read complete\n");
#endif
		break;
	case MODE_float:
		cdata=(unsigned char *)rdata;
		flags&=~MRC_COMPLEX;
		for (j=0; j<ny; j++) {
			if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,in))!=nz) {
				printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
				return 0;
			}
			do_progress(j*nx*4,ny*nx*4,"Reading float data...");
		}
		// swap order if machine and file have different order
		if (ord+mord==1) {
			do_progress(ny*nx*4,ny*nx*4,"Swapping byte order");
			for (i=nx*ny*nz*4-4; i>=0; i-=4) {
				c=cdata[i]; cdata[i]=cdata[i+3]; cdata[i+3]=c;
				c=cdata[i+1]; cdata[i+1]=cdata[i+2]; cdata[i+2]=c;
			}
		}
		break;
	case MODE_short_COMPLEX:
		cdata=(unsigned char *)rdata;
		flags|=MRC_COMPLEX;
		for (j=0; j<ny; j++) {
			if ((i=fread(&cdata[j*nx*nz*2],nx*sizeof(short),nz,in))!=nz) {
				printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
				return 0;
			}
			do_progress(j*nx*2,ny*nx*2,"Reading complex 16 bit data...");
		}
		do_progress(ny*nx*2,ny*nx*2,"Converting to float");
		if (ord) {
			for (i=nx*ny*nz-1; i>=0; i--) {
				rdata[i]=(float)((cdata[i*2]<<8)+cdata[i*2+1]); 
			}
		}
		else {
			for (i=nx*ny*nz-1; i>=0; i--) {
				rdata[i]=(float)((cdata[i*2+1]<<8)+cdata[i*2]); 
			}
		}
		break;
	case MODE_float_COMPLEX:
		cdata=(unsigned char *)rdata;
		flags|=MRC_COMPLEX;
		clearerr(in);
		for (j=0; j<ny; j++) {
			if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,in))!=nz) {
				printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
				return 0;
			}
			do_progress(j*nx*4,ny*nx*4,"Reading complex float data...");
		}
		if (ord+mord==1) {
			do_progress(ny*nx*4,ny*nx*4,"Swapping byte order");
			for (i=0; i<nx*ny*nz; i++) {
				w=cdata[i*4+3];
				cdata[i*4+3]=cdata[i*4];
				cdata[i*4]=w;
				w=cdata[i*4+2];
				cdata[i*4+2]=cdata[i*4+1];
				cdata[i*4+1]=w;
			}
		}
		if (ferror(in)) return 0;
		break;
	default: printf("Bad MRC file\n"); return 0;
	}
//}

if (flags&MRC_COMPLEX) {
	for (i=0; i<nx*ny*nz; i+=2) {
		a=hypot(rdata[i],rdata[i+1])/nx*ny*nz;
		p=atan2(rdata[i+1],rdata[i]);
		rdata[i]=a;
		rdata[i+1]=p;
	}
}


//pixres=header.xlen;
//if (pixres<=0 || pixres==1.0) pixres=25.0/3.0;
//mag=pixres*90.0/25.0;

mag=getenvfd("mag",30.0);
pixres=getenvfd("scanres",25.0)/mag*10.0;

do_progress(1,1,"Updating statistics");
update();
do_progress(-1,0,NULL);
//renMin=min;
//renMax=max;
if (pipe) pclose(in); else fclose(in);
return 1;
}

int MRCData:: readPGM( char * fsp)
{
FILE *in;
int i,j,l,pipe=0;
float f,a,p,div;
char s[800],*m2,w;
unsigned char *cdata,*cdata2,c;

#ifdef EECH
printf("PGM open '%s'\n",fsp);
#endif

in=fopen(fsp,"r");
if (in==NULL) { printf("Cannot open file %s\n",fsp); return 0; }
if (getc(in)!='P'||getc(in)!='5') {
	fclose(in);
	pipe=1;
	m2=fsp+strlen(fsp)-4;
	s[0]=0;
	if (!strcmp(m2,".gif") || !strcmp(m2,".GIF")) 
		sprintf(s,"giftoppm %s|ppmtopgm",fsp);
	else if (!strcmp(m2,".tif") || !strcmp(m2,".TIF") || !strcmp(m2-1,".tiff") )
		sprintf(s,"tifftopnm %s | ppmtopgm",fsp);
	else if (!strcmp(m2,".jpg") || !strcmp(m2,".JPG")) 
		sprintf(s,"djpeg -grayscale -pnm %s",fsp);
	else if (!strcmp(m2,".pcx") || !strcmp(m2,".PCX"))
		sprintf(s,"pcxtoppm %s| ppmtopgm",fsp);
	else sprintf(s,"ppmtopgm %s",fsp);

	in=popen(s,"r");
	if (getc(in)!='P'||getc(in)!='5') {
		pclose(in);
		printf("Sorry, I do not recognize this file type.\n");
		return 0;
	}
}

name=(char *)malloc(strlen(strrchr(fsp,'/'))+1);
strcpy(name,strrchr(fsp,'/')+1);
//setFile(name);
setFile(fsp);

flags=0;

nx=ny=0;
nz=1;
getc(in);
while (fgets(s,100,in)!=NULL && s[0]=='#') ;
sscanf(s," %d %d %f",&nx,&ny,&div);
getc(in);
if (nx<=0||ny<=0) { printf("Invalid file\n"); return 0; }

#ifdef EECH
printf("Image is %dx%d\n",nx,ny);
#endif

rdata=(float *)realloc(rdata,sizeof(float)*nx*ny*nz);
cdata=(unsigned char *)rdata;

for (j=0; j<ny; j++) {
	if ((i=fread(&cdata[(ny-j-1)*nx*nz],nx,nz,in))!=nz) {
		printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
		return 0;
	}
	do_progress(j*nx,nx*ny,"Reading 8 bit data...");
}
do_progress(nx*ny,nx*ny,"Converting to float");
printf("div=%f\n",div);
if (div<1.0 || div>=65536.0) div=255.0;
for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=(float)cdata[i]/div;

for (i=nx*ny*nz-1; i>=0; i--) if (rdata[i]>1.0||rdata[i]<0) printf("%d,%g\n",i,rdata[i]);

mag=getenvfd("mag",30.0);
pixres=getenvfd("scanres",25.0)/mag*10.0;

do_progress(1,1,"Updating statistics"	);
update();
do_progress(-1,0,NULL);
//renMin=min;
//renMax=max;
if (pipe) pclose(in); else fclose(in);
return 1;
}

int MRCData::readSAL( char * fsp)
{
char *fsp1;
FILE *in;
char s[81],s2[10],*t;
int i,j,rl=512,bl,sm=0;
unsigned char *cdata;
short *sdata,sw;

fsp1=(char *)malloc(strlen(fsp)+5);
strcpy(fsp1,fsp);
if (strrchr(fsp1,'.')!=NULL)  *((char *)strrchr(fsp1,'.'))=0;
strcat(fsp1,".HDR");
if (access(fsp1,R_OK)) strcpy(fsp1+strlen(fsp1)-4,".hdr");

if (strrchr(fsp,'/')==NULL) {
	setName("SAL File");
}
else {
	name=(char *)malloc(strlen(strrchr(fsp,'/'))+1);
	strcpy(name,strrchr(fsp,'/')+1);
//	setFile(name);
	setFile(fsp);
}
nx=ny=-1;
nz=1;

mag=getenvfd("mag",30.0);
pixres=getenvfd("scanres",25.0)/mag*10.0;

i=0;
if ((in=fopen(fsp1,"r"))==NULL) { printf("Cannot open %s\n",fsp1); return 0; }
while (fgets(s,80,in)!=NULL) {
	if (s[0]==0) continue;
	if (strncmp(s+1,"NXP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&nx);
	}
	else if (strncmp(s+1,"NYP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&ny);
	}
	else if (strncmp(s+1,"AXSCAN",6)==0) {
		t=strrchr(s,'\'');
		if (t[-1]=='Y') i=1;
	}
	else if (strncmp(s+1,"FILE REC",8)==0) {
		sscanf(strchr(s,'=')+1," %d",&rl);
	}
	else if (strncmp(s+1,"SCAN TYPE",9)==0) {
		sscanf(strchr(s,'\'')+1," %s",s2);
		if (s2[0]=='R') sm=1;
	}
	else if (strncmp(s+1,"DELTAX",6)==0) {
		sscanf(strchr(s,'=')+1," %f",&pixres);
		pixres/=3.0;
	}
}	
fclose(in);

if (i) { i=nx; nx=ny; ny=i; }
printf("%s is %dx%d\n",fsp1,nx,ny);
rdata=(float *)realloc(rdata,nx*(ny+1)*sizeof(float));
cdata=(unsigned char *)rdata;
sdata=(short *)rdata;
if (nx==-1 || ny==-1) { printf("Problem reading header."); return 0; }

*((char *)strrchr(fsp1,'.'))=0;
strcat(fsp1,".IMG");
if (access(fsp1,R_OK)) strcpy(fsp1+strlen(fsp1)-4,".img");
if ((in=fopen(fsp1,"r"))==NULL) { printf("Cannot open %s\n",fsp1); return 0; }

bl=(((nx*sizeof(short)-1)/rl)+1)*rl;
printf("Block length=%d x %d\n",bl,ny);
for (j=0; j<ny; j++) {
	if ((i=fread(&cdata[j*nx*nz*2],bl,nz,in))!=nz) {
		printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
		return 0;
	}
//	printf("Line %d is %d bytes\n",j,(((nx*sizeof(short)-1)/rl)+1)*rl);
	do_progress(j*nx*2,ny*nx*2,"Reading 16 bit data...");
}

if (ny*nx*4>40000) do_progress(ny*nx*2,ny*nx*2,"Converting to float");

// This reverses odd numbered rows in RASTER scan mode
if (sm) {
//	printf("Flipping data\n");
	for (i=1; i<ny; i+=2) {
		for (j=0; j<nx/2; j++) {
			sw=sdata[i*nx+j];
			sdata[i*nx+j]=sdata[i*nx+nx-j-1];
			sdata[i*nx+nx-j-1]=sw;
//			if (i==1) printf("%d<->%d\n",i*nx+j,i*nx+nx-j-1);
		}
	}
}

// Make the data float
for (i=nx*ny*nz-1; i>=0; i--)  
	rdata[i]=((float)((cdata[i*2+1]*256.0)+cdata[i*2])); 

//printf("%04x %04x %04x\n",(int)(rdata[0]*1000.0),(int)(rdata[nx]*1000.0),
//	(int)(rdata[nx*2]*1000.0));

update();
//renMin=min;
//renMax=max;
do_progress(-1,0,NULL);

fclose(in);
free(fsp1);
return 1;
}

int MRCData::readSALbox(char * fsp)
{
char *fsp1;
FILE *in;
char s[81],s2[10],*t;
int i,j,k,l,rl=512,bl,sm=0;
unsigned char *cdata;
short *sdata,sw;
int dx,dy,x0=0,x1=1040,y0=0,y1=1040;
short *blbuf;

fsp1=(char *)malloc(strlen(fsp)+5);
strcpy(fsp1,fsp);
if (strrchr(fsp1,'.')!=NULL)  *((char *)strrchr(fsp1,'.'))=0;
strcat(fsp1,".HDR");
if (access(fsp1,R_OK)) strcpy(fsp1+strlen(fsp1)-4,".hdr");

if (strrchr(fsp,'/')==NULL) {
	setName("SAL File");
}
else {
	name=(char *)malloc(strlen(strrchr(fsp,'/'))+1);
	strcpy(name,strrchr(fsp,'/')+1);
//	setFile(name);
	setFile(fsp);
}
nx=ny=-1;
nz=1;


mag=getenvfd("mag",30.0);
pixres=getenvfd("scanres",25.0)/mag*10.0;

i=0;
if ((in=fopen(fsp1,"r"))==NULL) { printf("Cannot open %s\n",fsp1); return 0; }
while (fgets(s,80,in)!=NULL) {
	if (s[0]==0) continue;
	if (strncmp(s+1,"NXP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&nx);
	}
	else if (strncmp(s+1,"NYP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&ny);
	}
	else if (strncmp(s+1,"AXSCAN",6)==0) {
		t=strrchr(s,'\'');
		if (t[-1]=='Y') i=1;
	}
	else if (strncmp(s+1,"FILE REC",8)==0) {
		sscanf(strchr(s,'=')+1," %d",&rl);
	}
	else if (strncmp(s+1,"SCAN TYPE",9)==0) {
		sscanf(strchr(s,'\'')+1," %s",s2);
		if (s2[0]=='R') sm=1;
	}
	else if (strncmp(s+1,"DELTAX",6)==0) {
		sscanf(strchr(s,'=')+1," %f",&pixres);
		pixres/=3.0;
	}
}	
fclose(in);

if (i) { i=nx; nx=ny; ny=i; }
printf("%s is %dx%d\n",fsp1,nx,ny);
//printf("Enter box (x,y,w,h): ");
//sscanf(fl_show_input("Enter box (x,y,w,h): ","")," %d , %d , %d , %d", &x0,&y0,&x1,&y1);
x1+=x0;
y1+=y0;
if (x1>nx || x1<0) x1=nx-1;
if (y1>ny || y1<0) y1=ny-1;
dx=x1-x0;
dy=y1-y0;
rdata=(float *)realloc(rdata,dx*dy*sizeof(float));
sdata=(short *)rdata;
if (nx==-1 || ny==-1) { printf("Problem reading header."); return 0; }

*((char *)strrchr(fsp1,'.'))=0;
strcat(fsp1,".IMG");
if (access(fsp1,R_OK)) strcpy(fsp1+strlen(fsp1)-4,".img");
if ((in=fopen(fsp1,"r"))==NULL) { printf("Cannot open %s\n",fsp1); return 0; }

bl=(((nx*sizeof(short)-1)/rl)+1)*rl;
if (sm) printf("Swapping\n");
	
printf("Block length=%d x %d (%d,%d %d,%d)\n",bl,ny,x0,y0,dx,dy);
blbuf=(short *)malloc(bl);
cdata=(unsigned char *)blbuf;
if (fseek(in,bl*y0,SEEK_SET)) printf("Seek error!!!\n");
for (j=y0; j<dy+y0; j++) {
	if (fread(blbuf,bl,1,in)!=1) {
		printf("Incomplete data read %d/%d blocks\n",j,ny*nz);
		return 0;
	}

	if (sm && (j&1)) {
		for (i=0; i<bl/4; i++) { 
			sw=blbuf[bl/2-i-1]; 
			blbuf[bl/2-i-1]=blbuf[i]; 
			blbuf[i]=sw;
		}
	}
	
	for (i=x0,k=(j-y0)*dx; i<=x1; i++,k++) {
		rdata[k]=((float)((cdata[i*2+1]*256.0)+cdata[i*2]));
	}
	do_progress(j*nx*2,ny*nx*2,"Reading 16 bit data...");
}

nx=dx;
ny=dy;
nz=1;
update();
//renMin=min;
//renMax=max;
do_progress(-1,0,NULL);

fclose(in);
free(fsp1);
free(blbuf);
return 1;
}

int MRCData::makeFFTAvg(List *list )
{
int i,j;
float *data,c,re,im;
char s[80];
MRCData *tmp;

if (list==NULL || list->count()<=1) return 0;
tmp=(MRCData *)list->objectAt(0);
if (tmp->xSize()!=npfar(tmp->xSize()) || tmp->ySize()!=npfar(tmp->ySize())) return 0;

tmp=((MRCData *)list->objectAt(0))->doFFT(0);
if (!tmp) return 0;
setSize(tmp->xSize(),tmp->ySize(),1);
flags|=MRC_COMPLEX;

bcopy((char *)tmp->getRealData(),(char *)rdata,nx*ny*nz*sizeof(float));
delete tmp;

for (i=1; i<list->count(); i++) {
		// the size check lags by one data set since we know i=0 is correct
	tmp=((MRCData *)list->objectAt(i))->doFFT(0);
	if (!tmp) return 0;
	if (nx!=tmp->xSize()||ny!=tmp->ySize()) return 0;
	data=tmp->getRealData();
	for (j=0; j<nx*ny*nz; j++) rdata[j]+=data[j];
	delete tmp;
}
c=(float)list->count();
for (j=0; j<=nx*ny*nz; j++) rdata[j]/=c;

tmp=(MRCData *)list->objectAt(0);
//printf("Mode %d average of %d sets\n",mode,list->count());
sprintf(s,"Avg of %d, %s",list->count(),tmp->Name());
setName(s);
update();

return 1;
}

// Average a list of data sets together
// mode=0 -> real averaging, mode=1 -> complex coherent average
// mode=2 -> complex incoherent average (doesn't work)
int MRCData:: makeAverage(List *list , int mode)
{
int i,j;
float *data,c,re,im;
char s[80];
MRCData *tmp;

if (list==0 || list->count()<=1) return 0;
tmp=(MRCData *)list->objectAt(0);
setSize(tmp->xSize(),tmp->ySize(),1);
if (tmp->isComplex()) flags|=MRC_COMPLEX; else flags&=~MRC_COMPLEX;

bcopy((char *)tmp->getRealData(),(char *)rdata,nx*ny*nz*sizeof(float));

switch (mode) {
case 0:
	for (i=1; i<list->count(); i++) {
		// the size check lags by one data set since we know i=0 is correct
		tmp=(MRCData *)list->objectAt(i);
		if (nx!=tmp->xSize()||ny!=tmp->ySize()) return 0;
		data=tmp->getRealData();
		for (j=0; j<nx*ny*nz; j++) {
			rdata[j]+=data[j];
//			if (fabs(rdata[j]>1.0e8)) printf("Bad %d %d %f %f\n",i,j,rdata[j],data[j]);
		}
	}
	c=(float)list->count();
	for (j=0; j<=nx*ny*nz; j++) rdata[j]/=c;
	break;
case 1:
	if (!(flags&MRC_COMPLEX)) return 0;
	// turn amp,pha -> re,im
	for (j=nx*ny*nz-2; j>=0; j-=2) {
		re=rdata[j]*cos(rdata[j+1]);
		im=rdata[j]*sin(rdata[j+1]);
		rdata[j]=re;
		rdata[j+1]=im;
	}

	for (i=1; i<list->count(); i++) {
		tmp=(MRCData *)list->objectAt(i);
		if (nx!=tmp->xSize()||ny!=tmp->ySize()) return 0;

		data=tmp->getRealData();
		for (j=nx*ny*nz-2; j>=0; j-=2) {
			rdata[j]+=data[j]*cos(data[j+1]);
			rdata[j+1]+=data[j]*sin(data[j+1]);
		}
		
	}

	c=list->count();
	// turn re,im -> amp,pha
	for (j=nx*ny*nz-2; j>=0; j-=2) {
		re=hypot(rdata[j]/c,rdata[j+1]/c);
		im=atan2(rdata[j+1]/c,rdata[j]/c);
		rdata[j]=re;
		rdata[j+1]=im;
	}
	break;
case 2:
	printf("Unimplimented\n");
	break;
}

tmp=(MRCData *)list->objectAt(0);
//printf("Mode %d average of %d sets\n",mode,list->count());
sprintf(s,"Avg of %d, %s",list->count(),tmp->Name());
setName(s);
update();

return 1;
}

int MRCData::setSize( int x , int y , int z)
{
rdata=(float *)realloc(rdata,x*y*z*sizeof(float));
nx=x;
ny=y;
nz=z;

return 1;
}

int MRCData::hasPstate()
{
if (pstate) return 1;
return 0;
}

LC_pstate *MRCData::Pstate()
{
if (!pstate) pstate=(LC_pstate *)malloc(sizeof(LC_pstate));
return pstate;
}

// NOTE: this flips the y axis as it reads (for imagic)
int MRCData::setFloatRealFromStream( FILE * strm , int x , int y , int z , int ord)
{
int i,j;
unsigned char *rd,t;

rdata=(float *)realloc(rdata,x*y*z*sizeof(float));
nx=x;
ny=y;
nz=z;

for (j=ny-1; j>=0; j--) {
	if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,strm))!=nz) {
		printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
		return 0;
	}
	if (nx>511) do_progress((ny-j)*nx*4,ny*nx*4,"Reading float data...");
}

// byte order reversal if appropriate
if (ord==1) {
	if (nx>511) do_progress(ny*nx*4,ny*nx*4,"Swapping byte order");
	brev((uchar *)rdata,nx*ny*nz);
}
update();
//renMin=min;
//renMax=max;
if (nx>511) do_progress(-1,0,NULL);
return 1;
}

// NOTE: this flips the y axis as it reads (for imagic)
int MRCData::setShortRealFromStream( FILE * strm , int x , int y , int z , int ord)
{
int i,j;
unsigned char *cdata;

// silly to undo this, but ...
i=1;
cdata=(uchar *)&i;
if (cdata[0]==0) ord--;

rdata=(float *)realloc(rdata,x*y*sizeof(float));
cdata=(unsigned char *)rdata;
nx=x;
ny=y;
nz=z;

for (j=ny-1; j>=0; j--) {
	if ((i=fread(&cdata[j*nx*nz*2],nx*sizeof(short),nz,strm))!=nz) {
		printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
		return 0;
	}
	if (nx>511) do_progress((ny-j)*nx*2,ny*nx*2,"Reading 16 bit data...");
}

if (nx>511) do_progress(ny*nx*2,ny*nx*2,"Converting to float");
if (ord==1) {
	for (i=nx*ny*nz-1; i>=0; i--)  
		rdata[i]=(float)((cdata[i*2]<<8)+cdata[i*2+1]); 
}
else {
	for (i=nx*ny*nz-1; i>=0; i--) 
		rdata[i]=(float)((cdata[i*2+1]<<8)+cdata[i*2]); 
}

update();
//renMin=min;
//renMax=max;
if (nx>511) do_progress(-1,0,NULL);
return 1;
}

int MRCData:: setRealData( float * buffer , int x , int y , int z)
{
int i;

flags&=~MRC_COMPLEX;
nx=x;
ny=y;
nz=z;
rdata=(float *)realloc(rdata,nx*ny*nz*sizeof(float));
bcopy((char *)buffer,(char *)rdata,nx*ny*nz*sizeof(float));
update();
return 1;
}

// remember, x*=2 for complex data
int MRCData:: setComplData( float * buffer , int x , int y , int z)
{
int i;

flags|=MRC_COMPLEX;
nx=x;
ny=y;
nz=z;
rdata=(float *)realloc(rdata,nx*ny*nz*sizeof(float));
bcopy((char *)buffer,(char *)rdata,nx*ny*nz*sizeof(float));
update();
return 1;
}

float * MRCData::getRealData() { return rdata; }
float * MRCData::getImagData() { return rdata+1; }
int MRCData::xSize() { return nx; }
int MRCData::ySize() { return ny; }
int MRCData::zSize() { return nz; }
int MRCData::Flags() { return flags; }
int MRCData:: setFlags( int f) { flags=f; return 1;}

// 2d files only
int MRCData:: extractCompl( float * buffer , int x0 , int y0 , int xsize , int ysize)
{
int i=0,x,y;

if (!(flags&MRC_COMPLEX)) return 0;
if (x0<0) x0=0;
if (y0<0) y0=0;
if (x0+xsize>nx) x0=nx-xsize;
if (y0+ysize>ny) y0=ny-ysize;

for (y=y0,i=0; y<y0+ysize; y++,i+=xsize) {
	bcopy((char *)&rdata[y*nx],(char *)&buffer[i],xsize*sizeof(float));
}

return 1;
}

// 2d files only
int MRCData:: extractReal( float * buffer , int x0 , int y0 , int xsize , int ysize)
{
int i=0,x,y;

//printf("Extract %d,%d %dx%d\n",x0,y0,xsize,ysize);
if (flags&MRC_COMPLEX) return 0;
if (x0<0) x0=0;
if (y0<0) y0=0;
if (x0+xsize>nx) x0=nx-xsize;
if (y0+ysize>ny) y0=ny-ysize;

for (y=y0,i=0; y<y0+ysize; y++,i+=xsize) {
	bcopy((char *)&rdata[y*nx+x0],(char *)&buffer[i],xsize*sizeof(float));
}

return 1;
}

MRCData *MRCData::doFFT(int ii)
{
MRCData *dat;
float *d,re,im,norm;
int i,j,k,l;
char s[80];
MRC_header *newh;

if (flags&MRC_COMPLEX) { printf("Cannot do FFT of FFT\n"); return NULL; }

if (nx!=npfa(nx) || ny!=npfa(ny)) { printf("Invalid FFT size!\n"); return NULL; }

dat=new MRCData(this);
dat->setSize(nx+2,ny,nz);
sprintf(s,"FFT %s",name);
dat->setName(s);

pfa2rc(1,1,nx,ny,rdata,(complex *)dat->getRealData());
pfa2cc(1,2,nx/2+1,ny,(complex *)dat->getRealData());

if (ii) dat->setFlags(MRC_COMPLEX|MRC_QSPACE|MRC_COMPLEXRI);
else dat->setFlags(MRC_COMPLEX|MRC_QSPACE);
d=dat->getRealData();

// (re,im) -> (amp,pha)
if (!ii) {
	norm=ny;
	for (i=0; i<(nx+2)*ny; i+=2) {
		re=hypot(d[i],d[i+1])/norm;
		im=atan2(d[i+1],d[i]);
		d[i]=re;
		d[i+1]=im;
	}
}

// This 'rotates' the data vertically by ny/2
l=ny/2*(nx+2);
for (i=0; i<ny/2; i++) {
	for (j=0; j<nx+2; j++) {
		k=j+i*(nx+2);
		re=d[k];
		d[k]=d[k+l];
		d[k+l]=re;
	}
} 

dat->update();
return dat;
}

MRCData *MRCData::doIFT(int ii)
{
MRCData *dat;
float *d,x,y,norm;
int i,j,k,l;
char s[80];

if (!(flags&MRC_COMPLEX)) { printf("Cannot do IFT of real\n"); return NULL; }

if (nx-2!=npfa(nx-2) || ny!=npfa(ny)) return NULL;

dat=new MRCData(this);
dat->setSize(nx,ny,nz);
sprintf(s,"IFT %s",name);
dat->setName(s);

d=dat->getRealData();
bcopy((char *)rdata,(char *)d,nx*ny*sizeof(float));

// This 'rotates' the data vertically by ny/2
l=ny/2*nx;
for (i=0; i<ny/2; i++) {
	for (j=0; j<nx; j++) {
		k=j+i*nx;
		x=d[k];
		d[k]=d[k+l];
		d[k+l]=x;
	}
} 

if (!ii) {
	norm=ny;
	for (i=nx*ny-2; i>=0; i-=2) {
		x=norm*d[i]*cos(d[i+1]);
		y=norm*d[i]*sin(d[i+1]);
		d[i]=x/(nx*ny);
		d[i+1]=y/(nx*ny);
	}
}

pfa2cc(-1,2,nx/2,ny,(complex *)d);
pfa2cr(-1,1,nx-2,ny,(complex *)d,d);

dat->setSize(nx-2,ny,nz);

dat->update();
return dat;
}

// This does an inv fft of the current data set into 'dest'. The current data set must
// be real,complex not amp,pha
void MRCData::doIFT(MRCData *dat)
{
float *d,x,y,norm;
int i,j,k,l;
char s[80];

if (!(flags&MRC_COMPLEX)) { printf("Cannot do IFT of real\n"); return; }

if (nx-2!=npfa(nx-2) || ny!=npfa(ny)) return;

dat->setSize(nx,ny,nz);
sprintf(s,"IFT %s",name);
dat->setName(s);

d=dat->getRealData();
bcopy((char *)rdata,(char *)d,nx*ny*sizeof(float));

// This 'rotates' the data vertically by ny/2
l=ny/2*nx;
for (i=0; i<ny/2; i++) {
        for (j=0; j<nx; j++) {
                k=j+i*nx;
                x=d[k];
                d[k]=d[k+l];
                d[k+l]=x;
        }
} 

pfa2cc(-1,2,nx/2,ny,(complex *)d);
pfa2cr(-1,1,nx-2,ny,(complex *)d,d);

dat->setSize(nx-2,ny,nz);

dat->update();
return ;
}

int MRCData::update()
{
long i,j,k,di;
double u,v,w;
double m,s,t;

#ifdef EECH
printf("Update %p, %d\n",this,nx*ny*nz);
#endif

max=-1.0e20;
min=-max;
m=0;
s=0;

for (i=0; i<256; i++) hist[i]=0;
if (flags&MRC_COMPLEX) di=2; else di=1;

// pass 1, statistics
for (i=nx*ny*nz-di; i>=0; i-=di) {
	v=rdata[i];
//	if (v>1e5) { printf(" %d,%d=%f\n",i%nx,i/nx,v); continue; }
	if (v>max) max=v;
	if (v<min) min=v;
	m+=v;
	s+=v*v;
}
mean=m*di/(double)(nx*ny*nz);
sigma=sqrt(s*di/(double)(nx*ny*nz)-mean*mean);

// this finds the average around the (1 pixel) edge of the image
edgemean=0;
for (i=0,j=(ny-1)*nx; i<nx; i++,j++) edgemean+=rdata[i]+rdata[j]; 
for (i=0, j=nx-1; i<nx*ny; i+=nx,j+=nx) edgemean+=rdata[i]+rdata[j];
edgemean/=(double)(nx*2+ny*2);

// this is the standard deviation (of the whole image) from the edgemean
edgesig=sqrt(s*di/(double)(nx*ny*nz)+edgemean*(edgemean-2.0*mean));

histmax=max;
histmin=min;

if (mean!=mean || sigma!=sigma) {
	printf("ERROR - Bad statistics on this data set!\n");
	printf("mean=%lf sigma=%lf edgemean=%lf edgesigma=%lf  %lf %lf\n",mean,sigma,edgemean,edgesig,s,m);
	printf("%f %f %f %f\n",rdata[0],rdata[nx*ny-1],rdata[nx*ny/2],rdata[nx*ny/3]);
	mean=edgemean=0;
	sigma=edgesig=1.0;
}

if (histmax-mean>sigma*3.0) histmax=mean+sigma*3.0;
if (mean-histmin>sigma*3.0) histmin=mean-sigma*3.0;
if (histmax==histmin) histmax+=.0001;
if (max==min) max+=.0001;

#ifdef EECH
printf("Update hist %f-%f real %f-%f sigma %f\n",histmin,histmax,min,max,sigma);
#endif

// pass 2, histogram
rehist(histmin ,histmax);

// The first time we're called, we set default values for renMin and renMax
if (renMin==-999.0) {
	t=-1;
	for (i=0; i<256; i++) if (hist[i]>t) { t=hist[i]; k=i; }		// find the histogram maximum

	// in spot scan images we usually want to focus on the second lump, not the first
	for (i=k; i<256 && hist[i]>t/10.0; i++);
	for (j=i; i<254; i++)  if (hist[i]>t/4.0&&hist[i+2]>t/4.0) break;		// yes, j=i

	// if we found a second large lump, focus in on it
	if (i!=254) {
		t=min;
		while ( i<256) { if (hist[i]>t) { t=hist[i]; k=i; } i++; }
	}

	// we need to ignore the central spike
	if (k==0) k=2;
	if (k>253) k=253;
	t=(hist[k+2]+hist[k-2])/2.0;

	// these values provide pretty good initial contrast settings
	for (i=k+1; i<256 && hist[i]>t/10.0; i++) ;
	renMax=histmin+(float)i*(histmax-histmin)/256.0;

	for (i=k; i>0 && hist[i]>t/10.0; i--) ;
	renMin=histmin+(float)i*(histmax-histmin)/256.0;

	if (renMax!=renMax || renMin!=renMin || fabs(renMax)>1.0e10) {
		renMin=-1.0;
		renMax=1.0;
	}
}

return 1;
}

// The histogram actually samples a pseudorandom subset of the data for large
// data sets (for speed)
int MRCData::rehist( float newmin , float newmax)
{
int i,j,k,p0,p1,di;
float u,v,w,norm=0;
static int prime[] = { 1,3,7,11,17,23,37,59,127,253,511 } ;

for (i=0; i<256; i++) hist[i]=0;

histmin=newmin;
histmax=newmax;

// decide on the sampling based on how big the data set is
if (nx*ny*nz<300000) p0=p1=0;
else if (nx*ny*nz<2000000) { p0=2; p1=3; }
else if (nx*ny*nz<8000000) { p0=4; p1=6; }
else { p0=7; p1=9; }

// real images don't need as much sampling
if (!(flags&MRC_COMPLEX)&&p0>0) { p0++; p1++; }

for (k=p0; k<=p1; k++) {
	if (flags&MRC_COMPLEX) di=prime[k]*2; else di=prime[k];
	norm+=nx*ny*nz/(float)di;
	w=255.999/(histmax-histmin);
	for (i=nx*ny*nz-di; i>=0; i-=di) {
		v=(rdata[i]-histmin)*w;	
		u=v-floor(v);
		j=(int)floor(v);
		if (j<0||j>255) continue;
		hist[j+1]+=u;
		hist[j]+=1.0-u;
	}
}

for (i=0; i<256; i++) hist[i]=hist[i]*100.0/norm;

return 1;
}

// this one has been somewhat optimized
int MRCData::renderAmp8( unsigned char * data , int x0 , int y0 , int ixsize , int iysize ,int bpl ,float scale ,int mingray ,int maxgray)
{
int i,j,k,l,ll,ymin=0,xmin,ymax,lmax,xsize,ysize;
int dsx,dsy; 	// used for integer scalings
int remx,remy,br,addi,addr;	// used for non integer scaling
float x,y,gs,t,rm;
unsigned char *tmp;

#ifdef EECH
printf("render %p %d,%d %d,%d bpl %d sca %4.2f %d-%d %f-%f\n",data,x0,y0,ixsize,iysize,bpl,
	scale,mingray,maxgray,renMin,renMax);
#endif

if (renMax==renMin) renMax-=.01;
rm=renMin;

scale=1.0/scale;
ysize=iysize;
xsize=ixsize;
if (iysize*scale>ny) ymin=(int)(iysize-ny/scale);
gs=(maxgray-mingray)/(renMax-renMin);
if (renMax<renMin) { gs=0;  rm=1.0e20; }
if (scale==floor(scale)) { dsx=(int)scale; dsy=(int)(scale*nx); } else dsx=-1;
if (dsx==-1) { remx=remy=0; addi=(int)floor(scale); addr=(int)(100000.0*(scale-floor(scale))); }
if (x0<0) { xmin=(int)(-x0/scale); xsize-=(int)floor(x0/scale); x0=0; } 
else { xmin=0; }
if ((xsize-xmin)*scale>(nx-x0)) xsize=(int)((nx-x0)/scale+xmin);
if (y0<0) { ymax=(int)(ysize+y0/scale-1); ymin+=(int)floor(y0/scale); y0=0; } else ymax=ysize-1;

if (xmin<0) xmin=0;
if (ymin<0) ymin=0;
if (xsize>ixsize) xsize=ixsize;
if (ymax>iysize) ymax=iysize;

lmax=nx*ny-1;

// complex version very similar to real version
if (flags&MRC_COMPLEX) {
// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			for (i=xmin,ll=x0; i<xsize; i++,ll+=dsx) {
				if (l+ll>lmax || ll>=nx-2) break;
	// +2 here is to prevent displaying qx=0 twice
				if (ll>=nx/2) {
					if (l>=(ny-scale)*nx) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;
				t=rdata[k];
				if (t<=rm) k=mingray;	// renMin, but deals with renMin>renMax
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
			}
			l+=dsy;
		}
	}
// non integer scalings a bit slower
	else {
		remy=10; // a small fixed fraction
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			br=l;
			remx=10;
			for (i=xmin,ll=x0; i<xsize-1; i++) {
				if (l+ll>lmax || ll>=nx-2) break;
				if (ll>=nx/2) {
					if (l>=(ny*nx-nx/scale)) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;

				t=rdata[k];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
				ll+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; ll++; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx; }
		}
	}
}
else {
// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=x0+y0*nx; j>=ymin; j--) {
			br=l;
			for (i=xmin; i<xsize; i++,l+=dsx) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
			}
			l=br+dsy;
		}
	}
// non integer scalings a bit slower
	else {
		remy=10;
		for (j=ymax,l=x0+y0*nx; j>=ymin; j--) {
			br=l;
			remx=10;
			for (i=xmin; i<xsize; i++) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
				l+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; l++; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx; }
		}
	}
}

return 1;
}

// this one has been somewhat optimized
int MRCData::renderPha8( unsigned char * data , int x0 , int y0 , int ixsize , int iysize ,int bpl ,float scale ,int mingray ,int maxgray)
{
int i,j,k,l,ll,ymin=0,xmin,ymax,lmax,xsize,ysize;
int dsx,dsy; 	// used for integer scalings
int remx,remy,br,addi,addr;	// used for non integer scaling
float x,y,gs,t;
unsigned char *tmp;

if (!(flags&MRC_COMPLEX)) 
	return renderAmp8(data,x0,y0,ixsize,iysize,bpl,scale,mingray,maxgray);

#ifdef EECH
printf("render %p %d,%d %d,%d bpl %d sca %4.2f %d-%d %f-%f\n",data,x0,y0,ixsize,iysize,bpl,
	scale,mingray,maxgray,renMin,renMax);
#endif

if (renMin==renMax) renMax+=.001;

scale=1.0/scale;
ysize=iysize;
xsize=ixsize;
if (iysize*scale>ny) ymin=(int)(iysize-ny/scale);
gs=(maxgray-mingray)/(2.0*M_PI);
if (scale==floor(scale)) { dsx=(int)scale; dsy=(int)(scale*nx); } else dsx=-1;
if (dsx==-1) { remx=remy=0; addi=(int)floor(scale); addr=(int)(100000.0*(scale-floor(scale))); }
if (x0<0) { xmin=(int)(-x0/scale); xsize-=(int)floor(x0/scale); x0=0; } 
else { xmin=0; }
if ((xsize-xmin)*scale>(nx-x0)) xsize=(int)((nx-x0)/scale+xmin);
if (y0<0) { ymax=(int)(ysize+y0/scale-1); ymin+=(int)floor(y0/scale); y0=0; } else ymax=ysize-1;

if (xmin<0) xmin=0;
if (ymin<0) ymin=0;
if (xsize>ixsize) xsize=ixsize;
if (ymax>iysize) ymax=iysize;

lmax=nx*ny-1;

// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			for (i=xmin,ll=x0; i<xsize; i++,ll+=dsx) {
				if (l+ll>lmax || ll>=nx-2) break;
	// +2 here is to prevent displaying qx=0 twice
				if (ll>=nx/2) {
					if (l>=(ny*nx-(float)nx/scale)) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;
				t=rdata[k+1];
				if (t<=-M_PI) k=mingray;
				else if (t>=M_PI) k=maxgray;
				else {
					k=(int)(gs*(t+M_PI));
					k+=mingray;
				}
				data[i+j*bpl]=k;
			}
			l+=dsy;
		}
	}
// non integer scalings a bit slower
	else {
		remy=10; // a small fixed fraction
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			br=l;
			remx=10;
			for (i=xmin,ll=x0; i<xsize-1; i++) {
				if (l+ll>lmax || ll>=nx-2) break;
				if (ll>=nx/2) {
					if (l>=(ny*nx-(float)nx/scale)) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;

				t=rdata[k+1];
				if (t<=-M_PI) k=mingray;
				else if (t>=M_PI) k=maxgray;
				else {
					k=(int)(gs*(t+M_PI));
					k+=mingray;
				}
				data[i+j*bpl]=k;
				ll+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; ll++; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx; }
		}
	}


return 1;
}

// NEEDS TO BE FIXED, DOES NOT WORK!!!!
int MRCData::renderAmp24( unsigned char * data , int x0 , int y0 , int ixsize , int iysize, int bpl ,float scale)
{
int i,j,k,l,ll,ymin=0,xmin,ymax,lmax,xsize,ysize;
int dsx,dsy; 	// used for integer scalings
int remx,remy,br,addi,addr;	// used for non integer scaling
float x,y,gs,t;
unsigned char *tmp;

scale=1.0/scale;
ysize=iysize;
if (ixsize*scale>nx) xsize=(int)(nx/scale); else xsize=ixsize;
if (iysize*scale>ny) ymin=(int)(iysize-ny/scale);
gs=256.0/(renMax-renMin);
if (scale==floor(scale)) { dsx=(int)(scale*2); dsy=(int)(2*scale*nx); } else dsx=-1;
if (dsx==-1) { remx=remy=0; addi=2*(int)floor(scale); addr=(int)(100000.0*(scale-floor(scale))); }
if (x0<0) { xmin=(int)(-x0/scale); xsize-=(int)floor(x0/scale); x0=0; } else xmin=0;
if (y0<0) { ymax=(int)(ysize+y0/scale-1); ymin+=(int)floor(y0/scale); y0=0; } else ymax=ysize-1;

if (xmin<0) xmin=0;
if (ymin<0) ymin=0;
if (xsize>ixsize) xsize=ixsize;
if (ymax>iysize) ymax=iysize;

lmax=nx*ny*2-1;

// complex version almost identical to real version
if (flags&MRC_COMPLEX) {
	if (dsx!=-1) {
		for (j=ymax,l=2*(x0+y0*nx); j>=ymin; j--) {
			br=l;
			for (i=xmin; i<xsize; i++,l+=dsx) {
				if (l>lmax) break;
				t=hypot(rdata[l],rdata[l+1]);
				if (t<=renMin) k=0;
				else if (t>=renMax) k=255;
				else {
					k=(int)(gs*(t-renMin));
				}
				ll=1+4*i+j*bpl;
				data[ll]=data[ll+1]=data[ll+2]=k;
			}
			l=br+dsy;
		}
	}
// non integer scalings a bit slower
	else {
		for (j=ymax,l=2*(x0+y0*nx); j>=ymin; j--) {
			br=l;
			remx=0;
			for (i=xmin; i<xsize; i++) {
				if (l>lmax) break;
				t=hypot(rdata[l],rdata[l+1]);
				if (t<=renMin) k=0;
				else if (t>=renMax) k=255;
				else {
					k=(int)(gs*(t-renMin));
				}
				ll=1+4*i+j*bpl;
				data[ll]=data[ll+1]=data[ll+2]=k;
				l+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; l+=2; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx*2; }
		}
	}
}
else {
// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=2*(x0+y0*nx); j>=ymin; j--) {
			br=l;
			for (i=xmin; i<xsize; i++,l+=dsx) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=renMin) k=0;
				else if (t>=renMax) k=255;
				else {
					k=(int)(gs*(t-renMin));
				}
				ll=1+4*i+j*bpl;
				data[ll]=data[ll+1]=data[ll+2]=k;
			}
			l=br+dsy;
		}
	}
// non integer scalings a bit slower
	else {
		for (j=ymax,l=2*(x0+y0*nx); j>=ymin; j--) {
			br=l;
			remx=0;
			for (i=xmin; i<xsize; i++) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=renMin) k=0;
				else if (t>=renMax) k=255;
				else {
					k=(int)(gs*(t-renMin));
				}
				ll=1+4*i+j*bpl;
				data[ll]=data[ll+1]=data[ll+2]=k;
				l+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; l+=2; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx*2; }
		}
	}
}

return 1;
}

float MRCData::Min() { return min; }
float MRCData::Max() { return max; }
float MRCData::Histmax() { return histmax; }
float MRCData::Histmin() { return histmin; }
float MRCData::Histscale() { return (histmax-mean)/(max-mean); }
float MRCData::Mean() { return mean; }
float MRCData::Sigma() { return sigma; }

void MRCData::normalize()
{
int i;

if (sigma==0||sigma!=sigma) return;

for (i=0; i<nx*ny*nz; i++) rdata[i]=(rdata[i]-mean)/sigma;

renMax=(renMax-mean)/sigma;
renMin=(renMin-mean)/sigma;

update();
}

void MRCData::edgeFlood(float val)
{
int i,j,k;

for (i=0; i<ny; i++) {
	for (j=0; j<nx; j++) {
		k=j+i*nx;
		if (rdata[k]<edgemean-edgesig/20.0 || rdata[k]>edgemean+edgesig/20.0) break;
		rdata[k]=val;
	}
	for (j=nx-1; j>=0; j--) {
		k=j+i*nx;
		if (rdata[k]<edgemean-edgesig/20.0 || rdata[k]>edgemean+edgesig/20.0) break;
		rdata[k]=val;
	}
}


for (j=0; j<nx; j++) {
	for (i=0; i<ny; i++) {
		k=j+i*nx;
		if (rdata[k]<edgemean-edgesig/20.0 || rdata[k]>edgemean+edgesig/20.0) break;
		rdata[k]=val;
	}
	for (i=ny-1; i>=0; i++) {
		k=j+i*nx;
		if (rdata[k]<edgemean-edgesig/20.0 || rdata[k]>edgemean+edgesig/20.0) break;
		rdata[k]=val;
	}
}

update();
}

void MRCData::edgeNormalize()
{
int i;

if (edgesig==0||edgesig!=edgesig) return;

for (i=0; i<nx*ny*nz; i++) rdata[i]=(rdata[i]-edgemean)/edgesig;

renMax=(renMax-edgemean)/edgesig;
renMin=(renMin-edgemean)/edgesig;

update();
}

int MRCData::setRenMinMax( float Min , float Max)
{
#ifdef EECH
printf("New minmax: %f %f\n",Min,Max);
#endif
renMin=Min;
renMax=Max;

return 1;
}

float MRCData::RenMin() { return renMin; }
float MRCData::RenMax() { return renMax; }
float * MRCData::Histogram() { return hist; }

float MRCData::Mag() { return mag; }
float MRCData::Pixres() { return pixres; }
int MRCData::setMag( float n) { mag=n; return 1; }
int MRCData::setPixres( float n) { pixres=n; return 1; }

int MRCData::clearBoxes()
{
nboxes=0;
return 1;
}

int MRCData::addBox( int x , int y , int w , int h,char *name)
{
if (nboxes>=boxesa-1) {
	boxesa+=8;
	boxes=(sRect *)realloc(boxes,boxesa*sizeof(sRect));
}
boxes[nboxes].x=x;
boxes[nboxes].y=y;
boxes[nboxes].width=w;
boxes[nboxes].height=h;
boxes[nboxes].x2=boxes[nboxes].y2=-1;
strncpy(boxes[nboxes].filespec,name,40);
boxes[nboxes].filespec[39]=0;
nboxes++;
return 1;
}

void MRCData::delBox(int boxno)
{
int i;

for (i=boxno; i<nboxes-1; i++) boxes[i]=boxes[i+1];
nboxes--;

return;
}

int MRCData::saveBoxDB( char * file)
{
int i;
FILE *out;

if (nboxes==0) { printf("No boxes to save!\n"); return 0; }
out=fopen(file,"w");
if (out==NULL) { printf("Unable to open file %s\n",file); return 0; }

fprintf(out,"!!! I.C.E. Datbase File !!!\n\n!!! Boxmrc Cutouts !!!\n");
for (i=0; i<nboxes; i++) {
	fprintf(out,"%hd %hd %hd %hd 0 %d 0 0 0 0 2 %s\n",boxes[i].x,boxes[i].y,boxes[i].width,boxes[i].height,boxes[i].height/2,boxes[i].filespec);
}
fprintf(out,"\n!!! Real Images !!!\n");
for (i=0; i<nboxes; i++) {
	fprintf(out,"%s\n",boxes[i].filespec);
}

fprintf(out,"!!! Fourier Images Complex !!!\n\n!!! Fourier Images Real !!!\n\n!!! Files with amplitudes and phases !!!\n\n");
fprintf(out,"!!! Spectra .com files  !!!\n\n!!! Miscellaneous Files !!!\n\n!!! Lattice Parameters !!!\n\n");

fclose(out);
return 1;
}

int MRCData::loadBoxDB( char * file)
{
FILE *in;
char s[120];
int i;

in=fopen(file,"r");
if (in==NULL) { printf("Unable to open file %s\n",file); return 0; }

while (fgets(s,119,in)!=NULL && strncmp(s,"!!! Boxmrc Cutouts",18)!=0);

while (fgets(s,119,in)!=NULL && strncmp(s,"!!!",3)!=0) {
	if (s[0]=='#') continue;
	
	if (sscanf(s," %hd %hd %hd %hd %d %d %d %d %d %d %d %s",&boxes[nboxes].x,&boxes[nboxes].y,
  &boxes[nboxes].width,&boxes[nboxes].height,&i,&i,&i,&i,&i,&i,&i,boxes[nboxes].filespec)==12) {
		boxes[nboxes].x2=boxes[nboxes].y2=-1;
		nboxes++;
		if (nboxes>=boxesa-1) {
			boxesa+=8;
			boxes=(sRect *)realloc(boxes,boxesa*sizeof(sRect));
    		}   
	}
}

fclose(in);
printf("%d boxes for this image\n",nboxes);
return 1;
}

int MRCData::isComplex()
{
if (flags&MRC_COMPLEX) return 1;
else return 0;
}

sRect *MRCData::Boxes() { return boxes; }
int MRCData::nBoxes() { return nboxes; }

float MRCData::clipToHist()
{
int i,di=1,nm=0;
float r;

if (flags&MRC_COMPLEX) di=2;

for (i=0; i<nx*ny*nz; i+=di) {
	if (rdata[i]>renMax) { rdata[i]=renMax; nm++; }
	if (rdata[i]<renMin) { rdata[i]=renMin; nm++; }
} 

update();

r=(float)nm/(float)(nx*ny*nz);
//printf("%5.3f%% modified\n",r);

return r;
}

int MRCData::setSeqno( int newno) { seqno=newno; return 1; }
int MRCData::Seqno() { return seqno; }

float MRCData::valueAt( int x , int y)
{
if (x>=nx||y>=ny||x<0||y<0) return 0;
return rdata[x+y*nx];
}

// type=0 is a step (circle) cutoff, type=1 is an atan() cutoff of ~ 4 pixels
int MRCData::applyMask( int type , int radius)
{
int i,j,k,di;
int max1,max2;
float r,v;

 
if (radius<=0) {
	if (nx<ny) max1=nx/2; else max1=ny/2;
	if (type) max1-=4;
}
else max1=(int)radius;
max2=(int)(((float)max1/sqrt(2.0))-1);

if (flags&MRC_COMPLEX) {
if (flags&MRC_COMPLEXRI) di=1; else di=2;
switch (type)
{
case 0:
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i+=di) {
			if (abs(j-ny/2)<max2 && i/2<max2) continue;
			r=hypot((float)(j-ny/2),(float)(i/2));		
			if (r>max1) rdata[i+j*nx]=0;
		}
	}
	break;
case 1:
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i+=di) {
			if (abs(j-ny/2)<max2 && i/2<max2) continue;
			r=hypot((float)(j-ny/2),(float)(i/2));		
			if (r>max1) {
				k=i+j*nx;
				// this falls off to ~96% at 4 pixels
				v=tanh(r-max1-2.0)/2.0+.5;
				rdata[k]=rdata[k]*(1.0-v);
			}
		}
	}
	break;
}
}
else {
switch (type)
{
case 0:
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			if (abs(j-ny/2+1)<max2 && abs(i-nx/2+1)<max2) continue;
			r=hypot((float)(j-(ny-1.0)/2.0),(float)(i-(nx-1.0)/2.0));		
			if (r+.5>max1) rdata[i+j*nx]=mean;
		}
	}
	break;
case 1:
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			if (abs(j-ny/2+1)<max2 && abs(i-nx/2+1)<max2) continue;
			r=hypot((float)(j-(ny-1.0)/2.0),(float)(i-(nx-1.0)/2.0));		
			if (r>max1) {
				k=i+j*nx;
				// this falls off to ~96% at 4 pixels
				v=tanh(r-max1-2.0)/2.0+.5;
				rdata[k]=mean*v+rdata[k]*(1.0-v);
			}
		}
	}
	break;
}
}

update();
return 1;
}

// This will multiply complex amplitudes by a radial function
// This can be used to apply synthetic CTF's to data
void MRCData::applyRadFn(int n, float *x, float *y)
{
int i, j, k=0, l;
float r, dx, x0;

x0=x[0];
dx=x[1]-x[0];

for (j=0; j<ny; j++) {
    for (i=0; i<nx; i+=2, k+=2) {
	r=hypot((float)(i/2), (float)(j-ny/2));
	r=(r-x0)/dx;
	l=(int)floor(r);
	r-=l;
	if (l>=n-2) rdata[k]*=y[n-1];
	if (l<0) rdata[k]*=y[0];
	else rdata[k]*=(y[l]*(1.0-r)+y[l+1]*r);
	if (rdata[k]<0) {
		rdata[k]=-rdata[k];
		if (rdata[k+1]<0) rdata[k+1]-=M_PI;
		else rdata[k+1]+=M_PI;
	}
    }
}
lastdist=-1;
update();
}

void MRCData::addRadNoise(int n, float *x, float *y,float nsig)
{
int i, j, k=0, l;
float r, f, dx, x0,re,im;
float r1,r2;

x0=x[0];
dx=x[1]-x[0];

for (j=0; j<ny; j++) {
    for (i=0; i<nx; i+=2, k+=2) {
	r=hypot((float)(i/2), (float)(j-ny/2));
	r=(r-x0)/dx;
	l=(int)floor(r);
	r-=l;
	if (l>=n-1) f=y[n-1];
	else if (l<0) f=y[0];
	else f=(y[l]*(1.0-r)+y[l+1]*r);
// gaussian distribution
//	f=grand(f,nsig);
	f=frand(f-nsig,f+nsig);
	re=rdata[k]*cos(rdata[k+1]);
	im=rdata[k]*sin(rdata[k+1]);
	re+=f*cos(r=frand(0,2.0*M_PI));
	im+=f*sin(r);
	rdata[k]=hypot(re,im);
	rdata[k+1]=atan2(im,re);
    }
}
lastdist=-1;
update();
}

// may not be very useful
void MRCData::subRadFn(int n, float *x, float *y)
{
int i, j, k=0, l;
float r, f, dx, x0;

x0=x[0];
dx=x[1]-x[0];

for (j=0; j<ny; j++) {
    for (i=0; i<nx; i+=2, k+=2) {
	r=hypot((float)(i/2), (float)(j-ny/2));
	r=(r-x0)/dx;
	l=(int)floor(r);
	r-=l;
	if (l>=n-1) f=y[n-1];
	if (l<0) f=y[0];
	else f=(y[l]*(1.0-r)+y[l+1]*r);
	rdata[k]-=f;
	if (rdata[k]<0) {
		rdata[k]=0;
		if (rdata[k+1]<0) rdata[k+1]-=M_PI;
		else rdata[k+1]+=M_PI;
	}
    }
}
update();
calcDist(2,10,10);
}

#define DISTN	200

// calculates various angular and radial averages of a complex data set
// Returns 1 pointer. 1st half of array is x 2nd half is y
float *MRCData::calcDist( int type , int x , int y)
{
int i,j,k,l;
float r,rr,t;

if (!(flags&MRC_COMPLEX)) return NULL;

if (!distdat) distdat=(float *)malloc(sizeof(float)*DISTN*2);

if ((lastdist==0 && type==0) || (lastdist==1 && type==1)) {
	for (i=0; i<DISTN; i++) distdat[i]=i*dr;
	return distdat;
}
lastdist=type;

bzero((char *)distdat,sizeof(float)*DISTN*2);
switch (type) {
case 0:			// average radial distribution
	dr=.6*(float)nx/(float)DISTN;
	for (i=0; i<nx; i+=2) {
		for (j=0; j<ny; j++) {
			r=hypot((float)(i>>1),(float)(j-ny/2));
			k=i+j*nx;
			rr=r/dr;
			l=(int)floor(rr);
			if (l>=DISTN-1) continue;
			rr-=(float)l;
			distdat[l]+=1.0-rr;
			distdat[l+1]+=rr;
			distdat[l+DISTN]+=rdata[k]*(1.0-rr);
			distdat[l+DISTN+1]+=rdata[k]*rr;
		}
	}
	break;
case 1:
	dr=M_PI/DISTN*2.0;
	for (i=0; i<nx; i+=2) {
		for (j=0; j<ny; j++) {
			r=hypot((float)(i>>1),(float)(j-ny/2));
			if (r>nx/2-2||r<3) continue;
			r=atan2((float)(j-ny/2),(float)(i/2));
			if (r<0) r+=M_PI;
			k=i+j*nx;
			rr=r/dr;
			l=(int)floor(rr);
			if (l>=DISTN-1) continue;
			rr-=(float)l;
//			if (l==0) printf("%f,%f\n",rr,rdata[k]);
			distdat[l]+=1.0-rr;
			distdat[l+1]+=rr;
			distdat[l+DISTN]+=rdata[k]*(1.0-rr);
			distdat[l+DISTN+1]+=rdata[k]*rr;
		}
	}
	// copy due to symmetry in the transform
	for (i=0; i<DISTN/2; i++) {
		distdat[i+DISTN/2]=distdat[i];
		distdat[i+3*DISTN/2]=distdat[i+DISTN];
	}
	break;
case 2:
	dr=.5*(float)nx/(float)DISTN;
	rr=atan2((float)y-ny/2,(float)x);
	for (r=0,k=0; r<(float)nx/2; r+=dr,k++) {
		for (t=rr-M_PI/18.0; t<rr+M_PI/18.0; t+=M_PI/360.0) {
			i=(int)(r*cos(t));
			j=(int)(r*sin(t));
			if (i<0) { j=-j; i=-i; }
			if (i>nx/2 || j>ny/2 || j<-ny/2) continue;
			i=2*i+(j+ny/2)*nx;
			distdat[k]++;
			distdat[k+DISTN]+=rdata[i];
		}
	}
			
	break;
case 3:
	dr=M_PI/DISTN*2.0;
	rr=hypot(x,y-ny/2);
	for (t=0,k=0; t<M_PI; t+=dr,k++) {
		for (r=rr-5; r<rr+5; r+=1.0) {
			i=(int)(r*cos(t));
			j=(int)(r*sin(t));
			if (i<0) { j=-j; i=-i; }
			if (i>nx/2 || j>ny/2 || j<-ny/2) continue;
			i=2*i+(j+ny/2)*nx;
			distdat[k]++;
			distdat[k+DISTN]+=rdata[i];
		}
	}
	// copy due to symmetry in the transform
	for (i=0; i<DISTN/2; i++) {
		distdat[i+DISTN/2]=distdat[i];
		distdat[i+3*DISTN/2]=distdat[i+DISTN];
	}
	break;
}

// normalize and set up x values
for (i=0; i<DISTN; i++) {
	if (distdat[i]>0) distdat[i+DISTN]/=distdat[i];
}

l=0;
for (i=1; i<DISTN; i++) {
	if (distdat[i]==0) {
		for (l=i+1; l<DISTN; l++) if (distdat[l]!=0) break;
		distdat[i+DISTN]=(distdat[l]+(float)(l-i)*distdat[i-1])/(float)(l-i+1);
	} 
	distdat[i]=i*dr;
}
distdat[0]=0;

//for (i=0; i<100; i++) printf("%f\t%f\n",distdat[i],distdat[i+DISTN]);
return distdat;
}

// Quite inefficient, but it's a start
void MRCData::calcRDist( float *dat, int RDISTN , int x )
{
int i,j,k,l,m;
float r,dr,dr2,t;

if (x<=0) return;

bzero((char *)dat,sizeof(float)*RDISTN*2);
dr=2.0*M_PI/RDISTN;
for (k=0; k<RDISTN; k++) {
	m=(int)(2*ceil(2.0*M_PI*(float)x/(float)RDISTN));
	if (m==0) m=2;
	dr2=dr/(float)m;
	for (t=k*dr-dr/2.0,l=0; l<=m; t+=dr2,l++) {
		for (r=x-2; r<x+3; r+=1.0) {
			i=(int)(r*cos(t));
			j=(int)(r*sin(t));
			if (i<-nx/2 || i>=nx/2 || j>=ny/2 || j<-ny/2) continue;
			i=i+nx/2+(j+ny/2)*nx;
			dat[k]++;
			dat[k+RDISTN]+=rdata[i];
		}
	}
}

// normalize and set up x values
for (i=0; i<RDISTN; i++) {
	if (dat[i]>0) dat[i+RDISTN]/=dat[i];
	else dat[i+RDISTN]=0;
	dat[i]=i*dr;
}

return; 
}

#define ROTSMP	1440

// Calculates a 'rotational correlation' between two data sets. If data
// set 2 were rotated by the returned angle the data sets would
// be rotationally aligned
float MRCData::calcRCF(MRCData *data2)
{
int i,j,nx2;
float dat[ROTSMP*2+2],dat2[ROTSMP*2+2],sum[ROTSMP];
MRCData *dd;
List *list;
float max,max2,norm;

if (!data2) { printf("Need a data set to align with!\n"); return 0; }

nx2=nx/2-10;

bzero((char *)sum,sizeof(float)*ROTSMP);
for (i=12; i<nx2; i+=6) {

	calcRDist(dat,ROTSMP,i);	// get the azimuthal distribution
	data2->calcRDist(dat2,ROTSMP,i);

	pfarc(-1,ROTSMP,&dat[ROTSMP],(complex *)&dat[ROTSMP]);	
			// do the fft
	pfarc(-1,ROTSMP,&dat2[ROTSMP],(complex *)&dat2[ROTSMP]);
	for (j=ROTSMP; j<ROTSMP*2+2; j+=2) {		
		max=dat[j]*dat2[j]+dat[j+1]*dat2[j+1];
		max2=dat[j+1]*dat2[j]-dat2[j+1]*dat[j];
		dat[j]=max;
		dat[j+1]=max2;
	}
	pfacr(1,ROTSMP,(complex *)&dat[ROTSMP],&dat[ROTSMP]);	// result of CCF
	norm=ROTSMP*ROTSMP*(fabs(mean)+sigma)*(fabs(data2->Mean())*data2->Sigma());
	for (j=0; j<ROTSMP; j++)
		sum[j]+=dat[j+ROTSMP]*(float)i/norm;
}

max2=1.0e30;
for (i=0; i<ROTSMP/2; i++) {
	max=sum[i]; sum[i]=sum[i+ROTSMP/2]; sum[i+ROTSMP/2]=max;
	max=dat[i]; dat[i]=dat[i+ROTSMP/2]-2.0*M_PI; dat[i+ROTSMP/2]=max;
	if (sum[i]<max2) max2=sum[i];
	if (sum[i+ROTSMP/2]<max2) max2=sum[i+ROTSMP/2];
}

for (i=0; i<ROTSMP; i++) sum[i]-=max2;

save_data(dat,sum,ROTSMP,"test.rcf.dat");

max=-1.0e25;
j=-1;
for (i=4; i<ROTSMP-4; i++) if (sum[i]>max) { max=sum[i]; j=i; }
printf("Max at %d, %g, %g\n",j,dat[j]/M_PI*180.0,max);

max=max2=0;
for (i=j-2; i<=j+2; i++) {
	max+=sum[i]*dat[i];
	max2+=sum[i];
}
printf("Max at %g\n",max/max2*180.0/M_PI);

return max/max2;
}

// Calculates the CCF with 'data2' or mirror ACF if data2 is NULL
MRCData *MRCData::calcCCF(MRCData *data2)
{
MRCData *f1,*f2,*cf;
float *rdata1,*rdata2;
int i,j,nx2;
float re,im,norm;

if (data2 && (data2->xSize()!=nx || data2->ySize()!=ny)) { printf("CCF: Images must be the same size.\n"); return NULL; }

f1=doFFT(1);
if (!f1) return NULL;
if (data2) f2=data2->doFFT(1); else f2=f1;
if (f2==NULL) return NULL;
rdata1=f1->getRealData();
rdata2=f2->getRealData();
nx2=nx+2;

if (data2) {    
	norm=nx2*ny*nx2*ny*(fabs(mean)+sigma)*(fabs(data2->Mean())+data2->Sigma());
	for (i=0; i<nx2*ny; i+=2) {
		re=(rdata1[i]*rdata2[i]+rdata1[i+1]*rdata2[i+1])/norm;
		im=(rdata1[i+1]*rdata2[i]-rdata1[i]*rdata2[i+1])/norm;
		rdata1[i]=re;
		rdata1[i+1]=im;
	}
}
else {                           
	norm=nx2*ny*nx2*ny*(mean+sigma)*(mean+sigma);
	for (i=0; i<nx2*ny; i+=2) {
		re=(rdata1[i]*rdata2[i]-rdata1[i+1]*rdata2[i+1])/norm;
		im=(rdata1[i+1]*rdata2[i]+rdata1[i]*rdata2[i+1])/norm;
		rdata1[i]=re;
		rdata1[i+1]=im;
	}
}
cf=f1->doIFT(1);
cf->setName("CCF");
cf->setFile("/tmp/emtool.ccf");

delete f1;
if (data2) delete f2;

return cf;
}

// Centers the current data set by doing an ACF, must already be +-10 pixels from being centered
// it will look for for negative peaks if it seems appropriate.
// if mode==1 it will center the image at the origin rather than the center
void MRCData::ACFcenter(int mode)
{
MRCData *f1,*f2,*cf;
float *data,*data2,norm;
int i,j,ii,jj,nx2,ifl=0;
float re,im,am,ph,f,to,peak[61][61];
char *tmp;

nx2=nx+2;
f1=doFFT(1);
if (!f1) return;
data=f1->getRealData();
f2=new MRCData(f1);
f2->setSize(nx2,ny,1);
f2->setFlags(MRC_COMPLEX);
data2=f2->getRealData();

norm=nx2*ny*nx2*ny*(mean+sigma)*(mean+sigma);
for (i=0; i<nx2*ny; i+=2) {
	re=(data[i]*data[i]-data[i+1]*data[i+1])/norm;
	im=(data[i+1]*data[i]+data[i]*data[i+1])/norm;
	data2[i]=re;
	data2[i+1]=im;
}

cf=f2->doIFT(1);
data=cf->getRealData();

printf("Center: ");
if (cf->Mean()-cf->Min()>cf->Max()-cf->Mean()) { ifl=1; printf("negative peak, "); }

// construct a small matrix (61x61) to search for a peak
for (i=0, ii=nx-31; i<61; i++,ii++) {
	if (ii==nx) ii=0;
	for (j=0, jj=ny-31; j<61; j++,jj++) {
		if (jj==ny) jj=0;
		if (ifl) peak[i][j]=-data[jj*nx+ii];
		else peak[i][j]=data[jj*nx+ii];
//putchar('0'+floor((peak[i][j]-cf->Min())/(cf->Max()-cf->Min())*9.99));
//if (ii==0 && jj==0) putchar('*');
	}
//putchar('\n');
}

// locate the peak
re=-1.0e20;
f=1.0e20;
ii=jj=0;
for (i=3; i<57; i++) {
	for (j=3; j<57; j++) {
		if (peak[i][j]>re) { re=peak[i][j]; ii=i; jj=j; }
		if (peak[i][j]<f) f=peak[i][j];
	}
}
printf("%d,%d  ",ii-30,jj-30);

// now find the peak to subpixel resolution
re=im=to=0;
for (i=ii-3; i<ii+4; i++) {
        for (j=jj-3; j<jj+4; j++) {
                re+=(float)i*(peak[i][j]-f);
                im+=(float)j*(peak[i][j]-f);
                to+=(peak[i][j]-f);
        }
}
re=re/to-30.0;
im=im/to-30.0;

printf("translate %4.6f,%4.6f\n",re,im); 
data=f1->getRealData();
re*=-.5;
im*=-.5;
if (mode) { re-=nx/2-.5; im-=ny/2-.5; }

// ok, now translate the image by doing the appropriate phase shifts in fourier space
for (i=-ny/2,ii=0; i<ny/2; i++) {
        for (j=0; j<nx2/2; j++,ii+=2) {
                am=hypot(data[ii],data[ii+1])/(float)(nx2*ny);
                ph=atan2(data[ii+1],data[ii]);
                ph+=2.0*M_PI*(re*(float)j/(float)(nx)+im*(float)i/(float)ny);
                data[ii]=am*cos(ph);
                data[ii+1]=am*sin(ph);
        }
}

tmp=name;
name=NULL;
f1->doIFT(this);
free(name);
name=tmp;
delete f1;
delete f2;
delete cf;
return;
}

// translates data2 so it is aligned with 'this', uses a CCF
// the data sets must be within 21 pixels of being aligned before this is called
// flag=1 looks for positive peaks, flag=-1 looks for negative peaks
// flag=0 will try to do whatever is appropriate
void MRCData::align(MRCData *data2,int flag)
{
MRCData *f1,*f2,*cf;
float *rdata1,*rdata2,*rdata3,norm;
int i,j,ii,jj,nx2;
float re,im,f,to,am,ph,peak[61][61];
char *tmp;

if (flag<-1) flag=-1;
if (flag>1) flag=1;
if (!data2) return;
if (data2->xSize()!=nx || data2->ySize()!=ny) { printf("Align: Images must be the same size.\n"); return; }

f1=doFFT(1);
f2=data2->doFFT(1); 
if (f1==NULL || f2==NULL) return;
rdata1=f1->getRealData();
rdata2=f2->getRealData();
nx2=nx+2;

norm=nx2*ny*nx2*ny*(mean+sigma)*(data2->Mean()+data2->Sigma());
for (i=0; i<nx2*ny; i+=2) {
	re=(rdata1[i]*rdata2[i]+rdata1[i+1]*rdata2[i+1])/norm;
	im=(rdata1[i+1]*rdata2[i]-rdata1[i]*rdata2[i+1])/norm;
	rdata1[i]=re;
	rdata1[i+1]=im;
}
cf=f1->doIFT(1);
rdata3=cf->getRealData();

printf("Align: "); 
if (!flag) {
	printf("%5.3f,%5.3f,%5.3f\t",cf->Max(),cf->Mean(),cf->Min());
	if (cf->Mean()-cf->Min()<cf->Max()-cf->Mean()) { flag=1; printf("positive CCF, "); }
	else { flag=-1; printf("negative CCF, "); }
}

// construct a small matrix (41x41) to search for a peak
for (i=0, ii=nx-30; i<61; i++,ii++) {
        if (ii==nx) ii=0;
        for (j=0, jj=ny-30; j<61; j++,jj++) {
                if (jj==ny) jj=0;
		peak[i][j]=(flag==1)?rdata3[jj*nx+ii]:-rdata3[jj*nx+ii];
        }
}

// locate the peak
re=-1.0e20;
f=1.0e20;
ii=jj=0;
for (i=3; i<57; i++) {
        for (j=3; j<57; j++) {
                if (peak[i][j]>re) { re=peak[i][j]; ii=i; jj=j; }
                if (peak[i][j]<f) f=peak[i][j];
        }
}
//printf("peak %f %d,%d\n",re,ii-30,jj-30);

// now find the peak to subpixel resolution
re=im=to=0;
for (i=ii-3; i<ii+4; i++) {
        for (j=jj-3; j<jj+4; j++) {
                re+=(float)i*(peak[i][j]-f);
                im+=(float)j*(peak[i][j]-f);
                to+=(peak[i][j]-f);
        }
}
re=re/to-30.0;
im=im/to-30.0;

printf("translate %4.6f,%4.6f\n",re,im); 

// ok, now translate the image by doing the appropriate phase shifts in fourier space
for (i=-ny/2,ii=0; i<ny/2; i++) {
        for (j=0; j<nx2/2; j++,ii+=2) {
                am=hypot(rdata2[ii],rdata2[ii+1])/(float)(nx2*ny);
                ph=atan2(rdata2[ii+1],rdata2[ii]);
                ph+=2.0*M_PI*(re*(float)j/(float)(nx)+im*(float)i/(float)ny);
                rdata2[ii]=am*cos(ph);
                rdata2[ii+1]=am*sin(ph);
        }
}

f2->doIFT(data2);
tmp=data2->Name();
for (i=7; i<=strlen(tmp); i++) tmp[i-7]=tmp[i];

delete f1;
delete f2;
delete cf;

return;
}

MRCData *MRCData::makeClip(int x0,int y0,int w,int h)
{
MRCData *dat;
float *d;
int i,j,k,l=0;

dat=new MRCData(this);
dat->setSize(w,h,1);
d=dat->getRealData();

for (j=y0; j<y0+h; j++) {
	k=j*nx+x0;	
	for (i=0; i<w; i++,k++,l++) {
 		d[l]=rdata[k];          
	}
}
                 
dat->update();
return dat;
}

void MRCData::flipData()
{
int i,j;
float t;

for (i=0; i<(ny+1)>>1; i++) {
	for (j=0; j<nx; j++) {
		t=rdata[i*nx+j];
		rdata[i*nx+j]=rdata[(ny-i-1)*nx+j];
		rdata[(ny-i-1)*nx+j]=t;
	}
}

update();
}

// 'this' is the data set to rotate. Result goes into 'dest'
// This does a 'fourier interpolation' to get nx*2 by ny*2 points for linear interpolation
MRCData *MRCData::rotateInto(float angle)
{
MRCData *intrp[4],*ftdata,*dest;
float *rdata2;
float *ddata;
float mx[2];
float *sdata[4];
float x,y,x2,y2,am,ph,t,u,p[4];
int i,j,k,ii,jj,kk,nx2,ud0,ud1;
float dx[3] = { -.5,0,.5 }, dy[3] = { 0, -.5,0 };

dest=new MRCData(this);
dest->setSize(nx,ny,1);
ddata=dest->getRealData();

/*ftdata=doFFT(1);
if (!ftdata) { printf("Bad fft size\n"); delete dest; return NULL; }
rdata2=ftdata->getRealData(); */

nx2=nx+2;
angle=angle/180.0*M_PI;

/*
for (i=0; i<nx2*ny; i++) rdata2[i]/=(nx2*ny);
//ftdata->applyMask(0,-1);		// this masks out a circle in fourier space.
intrp[3]=ftdata->doIFT(1);

// generates 3 fourier interpolated data sets
for (k=0; k<3; k++) {
	for (i=-ny/2,ii=0; i<ny/2; i++) {
        	for (j=0; j<nx2/2; j++,ii+=2) {
			am=hypot(rdata2[ii],rdata2[ii+1]);
                	ph=atan2(rdata2[ii+1],rdata2[ii]);
                	ph+=2.0*M_PI*(dx[k]*(float)j/(float)(nx)+dy[k]*(float)i/(float)ny);
                	rdata2[ii]=am*cos(ph);
                	rdata2[ii+1]=am*sin(ph);
        	}
	}
	intrp[k]=ftdata->doIFT(1);
}
//sdata[0]=rdata;
sdata[0]=intrp[3]->getRealData();
sdata[1]=intrp[0]->getRealData();
sdata[2]=intrp[1]->getRealData();
sdata[3]=intrp[2]->getRealData();
*/
sdata[0]=rdata;

mx[0]=cos(angle);
mx[1]=sin(angle);
for (j=0,y=-(ny-1.0)/2.0; j<ny; j++,y+=1.0) {
	for (i=0,x=-(nx-1.0)/2.0; i<nx; i++,x+=1.0) {
		// rotate coordinates
		x2=mx[0]*x+mx[1]*y+(nx-1.0)/2.0;
		y2=-mx[1]*x+mx[0]*y+(ny-1.0)/2.0;
		if (x2<0||x2>nx-1.0||y2<0||y2>ny-1.0) {
			ddata[i+j*nx]=mean;
			continue;
		}
		ii=(int)floor(x2);
		jj=(int)floor(y2);
		kk=ii+jj*nx;
		
		// no interpolation
		//ddata[i+j*nx]=sdata[0][ii+jj*nx];

		// bilinear interpolation of raw data (circular fourier mask)
		t=(x2-(float)ii);
		u=(y2-(float)jj);

		p[0]=sdata[0][kk];
		p[1]=sdata[0][kk+1];
		p[3]=sdata[0][kk+nx];
		p[2]=sdata[0][kk+nx+1];
		ddata[i+j*nx]=(1.0-t)*(1.0-u)*p[0]+t*(1.0-u)*p[1]+(1.0-t)*u*p[3]+t*u*p[2];

		// bilinear interpolation of fourier interpolated data
/*
		t=2.0*(x2-(float)ii);
		u=2.0*(y2-(float)jj);

		if (t<1.0&&u<1.0) {
			p[0]=sdata[0][kk];
			p[1]=sdata[1][kk];
			p[2]=sdata[2][kk];
			p[3]=sdata[3][kk];
		}
		else if (t<1.0) {
			p[0]=sdata[3][kk];
			p[1]=sdata[2][kk];
			p[2]=sdata[1][kk+nx];
			p[3]=sdata[0][kk+nx];
			u-=1.0;
		}
		else if (u<1.0) {
			p[0]=sdata[1][kk];
			p[1]=sdata[0][kk+1];
			p[2]=sdata[3][kk+1];
			p[3]=sdata[2][kk];
			t-=1.0;
		}
		else {
			p[0]=sdata[2][kk];
			p[1]=sdata[3][kk+1];
			p[2]=sdata[0][kk+nx+1];
			p[3]=sdata[1][kk+nx];
			t-=1.0;
			u-=1.0;
		}
		ddata[i+j*nx]=(1.0-t)*(1.0-u)*p[0]+t*(1.0-u)*p[1]+(1.0-t)*u*p[3]+t*u*p[2];
*/
	}
}

/*delete ftdata;
delete intrp[0];
delete intrp[1];
delete intrp[2];
delete intrp[3];*/

dest->update();
return dest;
}

// 'this' must be the re/im FFT of the data set. It will put the rotated result
// into 'dest'. The angle is in radians.
/*
void MRCData::rotateInto(MRCData *dest,float *angle)
{
float *ddata;
float x,y;
int i,j,k,ii,jj,nx2;

dest->setSize(nx-2,ny,1);
ddata=dest->getRealData();
nx2=nx-2;

for (ii=0,x=0; ii<nx2; ii++,x+=1.0) {
	for (jj=0,y=0; jj<ny; jj++,y+=1.0) {
		ddata[ii+jj*nx2]=0;
		for (i=-ny/2,k=0; i<ny/2; i++) {
        		for (j=0; j<nx2/2; j++,k+=2) {
				ddata[ii+jj*nx2]+=rdata[k]*cos(x*(float)j/(float)(nx2)+y*(float)i/(float)ny)-
					rdata[k+1]*sin(x*(float)j/(float)(nx2)+y*(float)i/(float)ny);
        		}
		}

	}
}

}
*/

void MRCData::subData(MRCData *img)
{
float *ddata;
int i;
float re,im;

if (nx!=img->xSize() || ny!=img->ySize()) { printf("Images must be the same size\n"); return; }

ddata=img->getRealData();

if (isComplex()) {
	for (i=0; i<nx*ny*nz; i+=2) {
		re=rdata[i]*cos(rdata[i+1])-ddata[i]*cos(ddata[i+1]);
		im=rdata[i]*sin(rdata[i+1])-ddata[i]*sin(ddata[i+1]);
		rdata[i]=hypot(re,im);
		rdata[i+1]=atan2(im,re);
	}
	update();
}
else {
	for (i=0; i<nx*ny*nz; i++) rdata[i]-=ddata[i];
	update();
}

}

// This will increase the size of an image and pad the edge
// with 0 -> the mean value from the image or 1-> random noise
void MRCData::padTo(int pad,int newx,int newy)
{
int oldx,oldy,padx,pady;
int i,j,ii,jj;

oldx=nx;
oldy=ny;
padx=newx-oldx;
pady=newy-oldy;
if (padx&1 || pady&1) { printf("Must pad an even number of pixels\n"); return; }
padx/=2;
pady/=2;

setSize(newx,newy,1);

for (i=(oldy-1)*oldx,ii=(oldy+pady-1)*newx; i>=0; i-=oldx,ii-=newx) {
	for (j=nx-1; j>=0; j--) {
		rdata[ii+j+padx]=rdata[i+j];
	}
}

for (j=pady*nx; j<(ny-pady)*nx; j+=nx) {
	for (i=0; i<padx; i++) {
		rdata[j+i]=pad?grand(mean,sigma):mean;
	}
	for (i=nx-padx; i<nx; i++) {
		rdata[j+i]=pad?grand(mean,sigma):mean;
	}
}

for (j=0; j<pady*nx; j++) {
	rdata[j]=pad?grand(mean,sigma):mean;
}

for (j=nx*(ny-pady); j<nx*ny; j++) {
	rdata[j]=pad?grand(mean,sigma):mean;
}

update();
}

#define COMMONRES	360

// This will do a search for common lines
MRCData *MRCData::findCommon(MRCData *data2)
{
MRCData *result;
float *rdata2,*rdatar;
float a1,a2,r,da=M_PI/(COMMONRES-1),t,u;
int i,j,k,ii,l1,l2,na;
float *cross1,*cross2;

na=COMMONRES*ny/2;
cross1=(float *)malloc(na*sizeof(float));
if (data2) cross2=(float *)malloc(na*sizeof(float));
else cross2=cross1;


result=new MRCData();
result->setSize(COMMONRES,COMMONRES,1);
rdatar=result->getRealData();
if (data2) rdata2=data2->getRealData();
else rdata2=rdata;

for (a1=-M_PI/2.0,j=i=0; j<COMMONRES; a1+=da,j++) {
	for (r=0,i=j*ny/2; r<ny/2-1; r+=1.0,i++) {
		l1=1+floor(r*cos(a1))*2+floor(r*sin(a1)+ny/2)*nx;
		cross1[i]=rdata[l1];
	}
}

if (data2) {
	for (a1=-M_PI/2.0,j=i=0; j<COMMONRES; a1+=da,j++) {
		for (r=0,i=j*ny/2; r<ny/2-1; r+=1.0,i++) {
			l1=1+floor(r*cos(a1))*2+floor(r*sin(a1)+ny/2)*nx;
			cross2[i]=rdata2[l1];
		}
	}
}

for (i=l1=0; i<na; i+=ny>>1) {
	for (j=0; j<na; j+=ny>>1,l1++) {
		rdatar[l1]=0;
		for (k=5; k<ny/2-5; k++) {
			rdatar[l1]+=asub(cross1[k+i],cross2[k+j]);
		}
	}
}

/*
for (a1=-M_PI/2.0,j=0; j<COMMONRES; a1+=da,j++) {
	for (a2=-M_PI/2.0,i=0; i<COMMONRES; a2+=da,i++) {
		k=i+j*COMMONRES;
		rdatar[k]=0;
		for (r=5.0; r<ny/2-1; r+=1.0) {
			l1=1+floor(r*cos(a1))*2+floor(r*sin(a1)+ny/2)*nx;
			l2=1+floor(r*cos(a2))*2+floor(r*sin(a2)+ny/2)*nx;

			t=asub(rdata[l1],rdata2[l2]);
			rdatar[k]+=t;
		}
	}
}
*/

free(cross1);
if (data2) free(cross2);

result->update();
return result;
}
