// MRCData.h - Copyright 1997  Steven Ludtke
// This is a generic EM data object extracted from 'emtool'
// It is not well documented at this point

#ifndef MRCDATAINC
#define MRCDATAINC

#include "List.h"
#include  "mrcP.h"
#include  "imagic.h"
#include <stdio.h>

typedef struct {
    short x, y;
    unsigned short width, height;
    short x2,y2;		// this is used for multiple image alignment
    char filespec[40];
} sRect;

typedef struct {
	float parm[8];	// 0 -> defocus, 1 -> gauss env, 2 -> scale, 3 -> ampcont
			// 4 -> gauss amp, 5 -> gauss wid, 6 -> exp amp, 7-> exp wid 
	float S;
} LC_pstate;

#define MRC_COMPLEX	1
#define MRC_QSPACE	2	// not really used consistently
#define MRC_COMPLEXRI	4	// this denotes a real/imag image (rather than Amp,Pha)

List *readIMAGIC(char *fsp);

class MRCData {
MRC_header *mrc_header;
IMAGICH *im_header;
int flags;
float *rdata;
int nx,ny,nz;
float min,max,histmax,histmin;
double mean,sigma,edgemean,edgesig;	// must be double or update() may mess up
float mag, pixres;
float hist[257];
float renMin,renMax;
char *name;
char *filespec;
float last_rmin,last_rmax;
sRect *boxes;
int boxesa,nboxes; 
int seqno;
float *distdat;
int lastdist;
float dr;   // used by calcdist
LC_pstate *pstate;

public:
MRCData();
MRCData(MRCData *src);
~MRCData();
int   update();
int   readSAL( char * file);
int   readSALbox(char * fsp);
int   readPGM( char * file);
int   readMRCData( char * file);
int   writeMRCData( char * fsp , int sh);
int   writePGM( char * fsp);
int   appendImagic( char * file);
int   makeFFTAvg(List *list);
int   makeAverage(List *list , int mode);
int   setFloatRealFromStream( FILE * strm , int x , int y , int z , int ord);
int   setShortRealFromStream( FILE * stream , int nx , int ny , int nz , int ord);
int   setRealData( float * buffer , int nx , int ny , int nz);
int   setComplData( float * buffer , int nx , int ny , int nz);
int   dataFromData(MRCData *src , int x0 , int y0 , int size);
int   setSize( int nx , int ny , int nz);
void setIMHed(IMAGICH *hed);
void setMRCHed(MRC_header *hed);
float *getRealData();
float *getImagData();
int   extractCompl( float * buffer , int x , int y , int xsize , int ysize);
int   extractReal( float * buffer , int x , int y , int xsize , int ysize);
MRCData *doFFT(int i);
MRCData *doIFT(int i);
void doIFT(MRCData *dest);
int   renderAmp8( unsigned char * data , int x , int y, int xsize, int ysize, int bpl, float scale, int mingray, int maxgray);
int   renderPha8( unsigned char * data , int x , int y, int xsize, int ysize, int bpl, float scale, int mingray, int maxgray);
// this one needs to be rewritten
int  renderAmp24( unsigned char * data , int x , int y , int xsize , int ysize , int bpl , float scale);
int  xSize();
int  ySize();
int  zSize();
int Flags();
int setFlags(int f);
float  Min();
float  Max();
float  Histmin();
float  Histmax();
float  Histscale();
float  Mean();
float  Sigma();
float  Mag();
float  Pixres();	    // pixel resolution (A/pix)
MRC_header *getHeader();
LC_pstate *Pstate();    
MRCData *makeClip(int x0,int y0,int w,int h);
int hasPstate();
int  setMag( float n);
int  setPixres(float n);
int  rehist( float newmin , float newmax);
int  setName( char * newname);
char *Name();
int  setFile( char * newname);
char *File();
float *Histogram();
int  setRenMinMax( float min , float max);
float  RenMin();
float  RenMax();
int  clearBoxes();
int  addBox( int x , int y , int w , int h,char *name);
int  saveBoxDB( char * file);
int  loadBoxDB( char * file);
int  isComplex();
sRect *  Boxes();
int  nBoxes();
void delBox(int boxno);
int  setSeqno( int newno);
int  Seqno();
float  valueAt( int x , int y);
float  clipToHist();
int  applyMask( int type , int r);
void normalize();
void edgeNormalize();
void edgeFlood(float val);
void applyRadFn(int n, float *x, float *y);
void addRadNoise(int n, float *x, float *y,float sigma);
void subRadFn(int n, float *x, float *y);
float *calcDist( int type , int x , int y);
float *Boxf();
void calcRDist( float *dat, int RDISTN, int x );
MRCData *calcCCF(MRCData *data2);
void ACFcenter(int );
void align(MRCData *data2,int flag);
MRCData *rotateInto(float angle);
float calcRCF(MRCData *data2);
void MRCData::flipData();
void subData(MRCData *img);
void padTo(int pad,int newx,int newy);
MRCData *findCommon(MRCData *data2);
};

inline MRC_header *MRCData::getHeader() { return mrc_header; }
#endif
