// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _complex_h
#ifdef __GNUG__
#pragma interface
#pragma cplusplus
#endif
#define _complex_h 1


#include <iostream.h>
extern "C" {
#include <math.h>
}
class complex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  float           re;
  float           im;

public:

  float           real() const;
  float           imag() const;

                   complex();
                   complex(const complex& y);
                   complex(float r, float i=0);

                  ~complex();

  complex&         operator =  (const complex& y);

  complex&         operator += (const complex& y);
  complex&         operator += (float y);
  complex&         operator -= (const complex& y);
  complex&         operator -= (float y);
  complex&         operator *= (const complex& y);
  complex&         operator *= (float y);

  complex&         operator /= (const complex& y); 
  complex&         operator /= (float y); 

  void             error(const char* msg) const;
};


// non-inline functions

complex   operator /  (const complex& x, const complex& y);
complex   operator /  (const complex& x, float y);
complex   operator /  (float   x, const complex& y);

complex   cos(const complex& x);
complex   sin(const complex& x);

complex   cosh(const complex& x);
complex   sinh(const complex& x);

complex   exp(const complex& x);
complex   log(const complex& x);

complex   pow(const complex& x, int p);
complex   pow(const complex& x, const complex& p);
complex   pow(const complex& x, float y);
complex   sqrt(const complex& x);
   
istream&  operator >> (istream& s, complex& x);
ostream&  operator << (ostream& s, const complex& x);

// other functions defined as inlines

int  operator == (const complex& x, const complex& y);
int  operator == (const complex& x, float y);
int  operator != (const complex& x, const complex& y);
int  operator != (const complex& x, float y);

complex  operator - (const complex& x);
complex  conj(const complex& x);
complex  operator + (const complex& x, const complex& y);
complex  operator + (const complex& x, float y);
complex  operator + (float x, const complex& y);
complex  operator - (const complex& x, const complex& y);
complex  operator - (const complex& x, float y);
complex  operator - (float x, const complex& y);
complex  operator * (const complex& x, const complex& y);
complex  operator * (const complex& x, float y);
complex  operator * (float x, const complex& y);

float  real(const complex& x);
float  imag(const complex& x);
float  abs(const complex& x);
float  norm(const complex& x);
float  arg(const complex& x);

complex  polar(float r, float t = 0.0);


// inline members

inline float  complex::real() const { return re; }
inline float  complex::imag() const { return im; }

inline complex::complex() {}
inline complex::complex(const complex& y) :re(y.real()), im(y.imag()) {}
inline complex::complex(float r, float i) :re(r), im(i) {}

inline complex::~complex() {}

inline complex&  complex::operator =  (const complex& y) 
{ 
  re = y.real(); im = y.imag(); return *this; 
} 

inline complex&  complex::operator += (const complex& y)
{ 
  re += y.real();  im += y.imag(); return *this; 
}

inline complex&  complex::operator += (float y)
{ 
  re += y; return *this; 
}

inline complex&  complex::operator -= (const complex& y)
{ 
  re -= y.real();  im -= y.imag(); return *this; 
}

inline complex&  complex::operator -= (float y)
{ 
  re -= y; return *this; 
}

inline complex&  complex::operator *= (const complex& y)
{  
  float r = re * y.real() - im * y.imag();
  im = re * y.imag() + im * y.real(); 
  re = r; 
  return *this; 
}

inline complex&  complex::operator *= (float y)
{  
  re *=  y; im *=  y; return *this; 
}


//  functions

inline int  operator == (const complex& x, const complex& y)
{
  return x.real() == y.real() && x.imag() == y.imag();
}

inline int  operator == (const complex& x, float y)
{
  return x.imag() == 0.0 && x.real() == y;
}

inline int  operator != (const complex& x, const complex& y)
{
  return x.real() != y.real() || x.imag() != y.imag();
}

inline int  operator != (const complex& x, float y)
{
  return x.imag() != 0.0 || x.real() != y;
}

inline complex  operator - (const complex& x)
{
  return complex(-x.real(), -x.imag());
}

inline complex  conj(const complex& x)
{
  return complex(x.real(), -x.imag());
}

inline complex  operator + (const complex& x, const complex& y)
{
  return complex(x.real() + y.real(), x.imag() + y.imag());
}

inline complex  operator + (const complex& x, float y)
{
  return complex(x.real() + y, x.imag());
}

inline complex  operator + (float x, const complex& y)
{
  return complex(x + y.real(), y.imag());
}

inline complex  operator - (const complex& x, const complex& y)
{
  return complex(x.real() - y.real(), x.imag() - y.imag());
}

inline complex  operator - (const complex& x, float y)
{
  return complex(x.real() - y, x.imag());
}

inline complex  operator - (float x, const complex& y)
{
  return complex(x - y.real(), -y.imag());
}

inline complex  operator * (const complex& x, const complex& y)
{
  return complex(x.real() * y.real() - x.imag() * y.imag(), 
                 x.real() * y.imag() + x.imag() * y.real());
}

inline complex  operator * (const complex& x, float y)
{
  return complex(x.real() * y, x.imag() * y);
}

inline complex  operator * (float x, const complex& y)
{
  return complex(x * y.real(), x * y.imag());
}

inline float  real(const complex& x)
{
  return x.real();
}

inline float  imag(const complex& x)
{
  return x.imag();
}

inline float  abs(const complex& x)
{
  return hypot(x.real(), x.imag());
}

inline float  norm(const complex& x)
{
  return (x.real() * x.real() + x.imag() * x.imag());
}

inline float  arg(const complex& x)
{
  return atan2(x.imag(), x.real());
}

inline complex  polar(float r, float t)
{
  return complex(r * cos(t), r * sin(t));
}

#endif
