// empgm.cc  Copyright 1997  Steve Ludtke
// v0.1 2/20/97
// Please read the A.README.1st file for Copyright information and
// usage restrictions

// Later versions will be commented more thouroughly

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "MRCData.h"
#include "List.h"
#include "cwp.h"

#define TYPE_NONE	0
#define TYPE_MRC	1	// MRC file format
#define TYPE_SAL	2	// ScansALot format
#define TYPE_IMAGIC	3	// Imagic file format


main(int argc, char *argv[])
{
int i,a,info=0,fft=0;
int type=TYPE_NONE;
MRCData *data;
List *list;
float min=1.0e10,max=1.0e10;

for (a=1; a<argc; a++) {
	if (argv[a][0]!='-') break;
	if (!strcmp(argv[a],"-mrc")) type=TYPE_MRC;
	else if (!strcmp(argv[a],"-sal")) type=TYPE_SAL;
	else if (!strcmp(argv[a],"-imagic")) type=TYPE_IMAGIC;
	else if (!strcmp(argv[a],"-fft")) fft=1;
	else if (!strcmp(argv[a],"-min")) { sscanf(argv[a+1]," %f",&min); a++; }
	else if (!strcmp(argv[a],"-max")) { sscanf(argv[a+1]," %f",&max); a++; }
	else if (!strcmp(argv[a],"-v")) info=1;
}

if (argc<=1 || a>argc-2 || type==TYPE_NONE) {
	printf("empgm v0.1, Copyright 1997 Steve Ludtke\n");
	printf("Converts MRC, ScansALot or Imagic files to PGM format\n\n");
	printf("empgm (-mrc -sal -imagic) [-fft] [-v] [-min <mingray>] [-max <maxgray>] <infile> <outfile>\n");
	exit(1);
}

switch(type) {
case TYPE_MRC:
	data=new MRCData;
	i=data->readMRCData(argv[a]);
	if (!i) { printf("Cannot read %s as a MRC file.\n",argv[a]); exit(1); }
	break;
case TYPE_SAL:
	data=new MRCData;
	i=data->readSAL(argv[a]);
	if (!i) { printf("Cannot read %s as a ScansALot file.\n",argv[a]); exit(1); }
	break;
case TYPE_IMAGIC:
	list=readIMAGIC(argv[a]);
	if (list->count()<1) { printf("Cannot read %s as an IMAGIC file.\n",argv[a]); exit(1); }
	data=(MRCData *)list->objectAt(0);
	delete list;
	break;
}

if (fft) {
	if (data->xSize()!=data->ySize()) { printf("Sorry, I can only do FFTs of square images.\n"); exit(0); }
	if (npfa(data->xSize())!=data->xSize() || npfa(data->ySize())!=data->ySize()) {
		printf("The image was not a valid size for the FFT routine.\nThe next valid size is %dx%d.\n",npfa(data->xSize()));
		printf("I have padded the image with noise to continue processing.\n");
		i=npfa(data->xSize());
		data->padTo(1,i,i);
	} 
	data=data->doFFT(0);
	if (!data) { printf("FFT unsuccessful.\n"); exit(1); }
}


if (info) {
	printf("\nThis image is %d x %d\n",data->xSize(),data->ySize());
	printf("greyscale:  min=%5.3g max=%5.3g mean=%5.3g sigma=%5.3g\n",data->Min(),data->Max(),
		data->Mean(),data->Sigma());
	printf("default greyscale limits %5.3g to %5.3g\n",data->RenMin(),data->RenMax());
}

if (max<1.0e10) data->setRenMinMax(data->RenMin(),max);
if (min<1.0e10) data->setRenMinMax(min,data->RenMax());

data->writePGM(argv[a+1]);
}

//
// The following routines are used by MRCData. 
// Most do not perform their original function here
//

void do_progress(int i,int ny, char *label)
{
int prog=0;
static long ltime = 0;
static char *llab=NULL;

if (i==-1) {
	if (prog) {
		printf("\n");
		prog=0;
	}
	ltime=0;
	llab=NULL;
	return;
}

if (!prog&&!ltime) ltime=time(0);

if (!prog) {
	ltime=time(0);
	llab=label;
	prog=1;
	return;
}

if (time(0)-ltime>0 || i==ny) {
	printf("Loading %d%%  \r",100*i/ny);
	ltime=time(0);
}

llab=label;
return;
}

// return a random number between lo and hi
float frand(float lo, float hi)
{
static int f=1;

if (f) { srandom(time(0)); f=0; }

return ((float)random()/2147483647.0*(hi-lo)+lo);
}

// return a gaussian distribution
float grand(float mean,float sigma)
{
float x,y,r,f;

do {
	x=frand(-1.0,1.0);
	y=frand(-1.0,1.0);
	r=x*x+y*y;
} while (r>1.0||r==0);
f=sqrt(-2.0*log(r)/r);

return x*f*sigma+mean;
}

float getenvfd(char *lab,float def) { return def; }

void save_data(float *x, float *y, int n, char *fsp)
{
FILE *out;
int i;

out=fopen(fsp,"w");
if (out==NULL) return;
for (i=0; i<n; i++) fprintf(out,"%g\t%g\n",x[i],y[i]);
fclose(out);
return;
}
